from collections import namedtuple
from ezdl.models.regseg import RegSeg48
from ezdl.models.backbones import ResNet, MiT
from ezdl.models.segnet import SegNet
from ezdl.models.dummy import Random, SingleConv
from ezdl.models.deeplab import deeplabv3_resnet50
from ezdl.models.lawin import Lawin, SplitLawin, DoubleLawin
from ezdl.models.base import WrappedModel
from super_gradients.training.models.kd_modules.kd_module import KDOutput

from ezdl.models.kd.feature import \
    FeatureDistillationModule, \
    FeatureDistillationConvAdapter, \
    VariationalInformationDistillation
    
from ezdl.models.kd.logits import LogitsDistillationModule

MODELS = {
    'conv': SingleConv,
    'resnet': ResNet,
    'mit': MiT,

    'segnet': SegNet,
    'regseg48': RegSeg48,
    'random': Random,
    'deeplabv3_resnet50': deeplabv3_resnet50,
    'lawin': Lawin,
    'splitlawin': SplitLawin,
    'doublelawin': DoubleLawin,
}

KD_MODELS = {
    "feature_distillation": FeatureDistillationModule,
    "variational_information_distillation": VariationalInformationDistillation,
    "feature_distillation_conv_adapter": FeatureDistillationConvAdapter,
    "logits_distillation": LogitsDistillationModule,
}


ComposedOutput = namedtuple('ComposedOutput', ['main',
                                    'aux'])


import importlib
from typing import Mapping

from super_gradients import KDTrainer
from super_gradients.common import StrictLoad
from ezdl.logger.text_logger import get_logger
from super_gradients.training.utils.checkpoint_utils import load_checkpoint_to_model
from super_gradients.training.models.all_architectures import KD_ARCHITECTURES

from ezdl.experiment.seg_trainer import SegmentationTrainer
from ezdl.models import KD_MODELS as KD_MODELS_DICT
from ezdl.utils.utilities import get_module_class_from_path

logger = get_logger(__name__)


class KDSegTrainer(SegmentationTrainer, KDTrainer):
    """
    Knowledge Distillation Trainer.
    """
    def init_model(self, params: Mapping, resume: bool, checkpoint_path: str = None):
        logger.info("Initializing teacher model")
        self.teacher_architecture, _ = self._load_model({**params, "model": params["kd"]["teacher"]})
        self.net = self.teacher_architecture
        self._net_to_device()
        self.teacher_architecture = self.net
        teacher_checkpoint = params.get("kd", {}).get("teacher", {}).get("checkpoint_path", None)
        if teacher_checkpoint is not None:
            load_checkpoint_to_model(params['kd']['teacher']['checkpoint_path'],
                                     load_backbone=False,
                                     net=self.teacher_architecture,
                                     strict=StrictLoad.ON.value,
                                     load_weights_only=False)
        else:
            logger.warning("No teacher checkpoint provided, using random weights")
        logger.info("Initializing student model")
        super().init_model(params, resume, checkpoint_path)

        self.net = self._load_kd_module(params['kd']['module'], student=self.net, teacher=self.teacher_architecture)

    def _save_best_checkpoint(self, epoch, state):
        """
        Overrides parent best_ckpt saving to modify the state dict so that we only save the student.
        """
        if self.ema:
            best_net = self.ema_model.ema.module.student
            state.pop("ema_net")
        else:
            best_net = self.net.module.student

        state["net"] = best_net.state_dict()
        self.sg_logger.add_checkpoint(tag=self.ckpt_best_name, state_dict=state, global_step=epoch)

    def _restore_best_params(self):
        self.checkpoint = load_checkpoint_to_model(
            ckpt_local_path=f'{self.checkpoints_dir_path}/ckpt_best.pth',
            load_backbone=False,
            net=self.net.module.student,
            strict=StrictLoad.ON.value,
            load_weights_only=True,
        )

    def _load_kd_module(self, model_params, student, teacher):
        arch_params = model_params['params']
        try:
            module, model_cls = get_module_class_from_path(model_params['name'])
            model_module = importlib.import_module(module)
            model_cls = getattr(model_module, model_cls)
            model = model_cls(arch_params, student=student, teacher=teacher)
        except (AttributeError, ValueError):
            if model_params['name'] in KD_MODELS_DICT.keys():
                model = KD_MODELS_DICT[model_params['name']](arch_params, student=student, teacher=teacher)
            else:
                model = KD_ARCHITECTURES[model_params['name']](arch_params, student=student, teacher=teacher)
        return model


"""
Code adapted from: https://github.com/Deci-AI/super-gradients
"""
import importlib
from typing import Union, Dict, Mapping, Any

import torch.nn
from omegaconf import DictConfig, OmegaConf
from torch.utils.data import DataLoader

from super_gradients.common import MultiGPUMode
from super_gradients.common.abstractions.abstract_logger import get_logger
from super_gradients.training import utils as core_utils, models
from super_gradients.training.dataloaders import dataloaders
from super_gradients.training.exceptions.kd_trainer_exceptions import (
    ArchitectureKwargsException,
    UnsupportedKDArchitectureException,
    InconsistentParamsException,
    UnsupportedKDModelArgException,
    TeacherKnowledgeException,
    UndefinedNumClassesException,
)
from super_gradients.training.models import SgModule
from super_gradients.training.models.all_architectures import KD_ARCHITECTURES
from super_gradients.training.models.kd_modules.kd_module import KDModule
from super_gradients.training.pretrained_models import PRETRAINED_NUM_CLASSES
from super_gradients.common import StrictLoad
from super_gradients.training.utils import get_param, HpmStruct
from super_gradients.training.utils.callbacks import KDModelMetricsUpdateCallback
from super_gradients.training.utils.checkpoint_utils import read_ckpt_state_dict, load_checkpoint_to_model
from super_gradients.training.utils.distributed_training_utils import setup_device
from super_gradients.training.utils.ema import KDModelEMA

from ezdl.models import KD_MODELS as KD_MODELS_DICT
from ezdl.utils.utilities import get_module_class_from_path
from ezdl.experiment.ez_trainer import EzTrainer
from ezdl.logger.text_logger import get_logger

logger = get_logger(__name__)


class KDEzTrainer(EzTrainer):
    """
    Knowledge Distillation Trainer.
    """
    def init_model(self, params: Mapping, resume: bool, checkpoint_path: str = None):
        logger.info("Initializing teacher model")
        self.teacher_architecture, _ = self._load_model({**params, "model": params["kd"]["teacher"]})
        self.net = self.teacher_architecture
        self._net_to_device()
        self.teacher_architecture = self.net
        teacher_checkpoint = params.get("kd", {}).get("teacher", {}).get("checkpoint_path", None)
        if teacher_checkpoint is not None:
            load_checkpoint_to_model(params['kd']['teacher']['checkpoint_path'],
                                     load_backbone=False,
                                     net=self.teacher_architecture,
                                     strict=StrictLoad.ON.value,
                                     load_weights_only=False)
        else:
            logger.warning("No teacher checkpoint provided, ensure you loaded them manually")
        logger.info("Initializing student model")
        super().init_model(params, resume, checkpoint_path)

        self.net = self._load_kd_module(params['kd']['module'], student=self.net, teacher=self.teacher_architecture)
        
    def print_model_summary(self, model=None):
        model = model or self.net
        logger.info("Student model summary:")
        return super().print_model_summary(model.student)

    def _save_best_checkpoint(self, epoch, state):
        """
        Overrides parent best_ckpt saving to modify the state dict so that we only save the student.
        """
        if self.ema:
            best_net = self.ema_model.ema.module.student
            state.pop("ema_net")
        else:
            best_net = self.net.module.student

        state["net"] = best_net.state_dict()
        self.sg_logger.add_checkpoint(tag=self.ckpt_best_name, state_dict=state, global_step=epoch)

    def _restore_best_params(self):
        self.checkpoint = load_checkpoint_to_model(
            ckpt_local_path=f'{self.checkpoints_dir_path}/ckpt_best.pth',
            load_backbone=False,
            net=self.net.module.student,
            strict=StrictLoad.ON.value,
            load_weights_only=True,
        )

    def _load_kd_module(self, model_params, student, teacher):
        arch_params = model_params['params']
        try:
            module, model_cls = get_module_class_from_path(model_params['name'])
            model_module = importlib.import_module(module)
            model_cls = getattr(model_module, model_cls)
            model = model_cls(arch_params, student=student, teacher=teacher)
        except (AttributeError, ValueError):
            if model_params['name'] in KD_MODELS_DICT.keys():
                model = KD_MODELS_DICT[model_params['name']](arch_params, student=student, teacher=teacher)
            else:
                model = KD_ARCHITECTURES[model_params['name']](arch_params, student=student, teacher=teacher)
        return model
    
    def _validate_args(self, arch_params, architecture, checkpoint_params, **kwargs):
        student_architecture = get_param(kwargs, "student_architecture")
        teacher_architecture = get_param(kwargs, "teacher_architecture")
        student_arch_params = get_param(kwargs, "student_arch_params")
        teacher_arch_params = get_param(kwargs, "teacher_arch_params")

        if get_param(checkpoint_params, "pretrained_weights") is not None:
            raise UnsupportedKDModelArgException("pretrained_weights", "checkpoint_params")

        if not isinstance(architecture, KDModule):
            if student_architecture is None or teacher_architecture is None:
                raise ArchitectureKwargsException()
            if architecture not in KD_ARCHITECTURES.keys():
                raise UnsupportedKDArchitectureException(architecture)

        # DERIVE NUMBER OF CLASSES FROM DATASET INTERFACE IF NOT SPECIFIED OR ARCH PARAMS FOR TEACHER AND STUDENT
        self._validate_num_classes(student_arch_params, teacher_arch_params)

        arch_params["num_classes"] = student_arch_params["num_classes"]

        # MAKE SURE TEACHER'S PRETRAINED NUM CLASSES EQUALS TO THE ONES BELONGING TO STUDENT AS WE CAN'T REPLACE
        # THE TEACHER'S HEAD
        teacher_pretrained_weights = core_utils.get_param(checkpoint_params, "teacher_pretrained_weights", default_val=None)
        if teacher_pretrained_weights is not None:
            teacher_pretrained_num_classes = PRETRAINED_NUM_CLASSES[teacher_pretrained_weights]
            if teacher_pretrained_num_classes != teacher_arch_params["num_classes"]:
                raise InconsistentParamsException(
                    "Pretrained dataset number of classes", "teacher's arch params", "number of classes", "student's number of classes"
                )

        teacher_checkpoint_path = get_param(checkpoint_params, "teacher_checkpoint_path")
        load_kd_model_checkpoint = get_param(checkpoint_params, "load_checkpoint")

        # CHECK THAT TEACHER NETWORK HOLDS KNOWLEDGE FOR THE STUDENT TO LEARN FROM OR THAT WE ARE LOADING AN ENTIRE KD
        if not (teacher_pretrained_weights or teacher_checkpoint_path or load_kd_model_checkpoint or isinstance(teacher_architecture, torch.nn.Module)):
            raise TeacherKnowledgeException()

    def _validate_num_classes(self, student_arch_params, teacher_arch_params):
        """
        Checks validity of num_classes for num_classes (i.e existence and consistency between subnets)

        :param student_arch_params: (dict) Architecture H.P. e.g.: block, num_blocks, num_classes, etc for student
        :param teacher_arch_params: (dict) Architecture H.P. e.g.: block, num_blocks, num_classes, etc for teacher

        """
        self._validate_subnet_num_classes(student_arch_params)
        self._validate_subnet_num_classes(teacher_arch_params)
        if teacher_arch_params["num_classes"] != student_arch_params["num_classes"]:
            raise InconsistentParamsException("num_classes", "student_arch_params", "num_classes", "teacher_arch_params")

    def _validate_subnet_num_classes(self, subnet_arch_params):
        """
        Derives num_classes in student_arch_params/teacher_arch_params from dataset interface or raises an error
         when none is given

        :param subnet_arch_params: Arch params for student/teacher

        """

        if "num_classes" not in subnet_arch_params.keys():
            if self.dataset_interface is None:
                raise UndefinedNumClassesException()
            else:
                subnet_arch_params["num_classes"] = len(self.classes)

    def _instantiate_net(self, architecture: Union[KDModule, KDModule.__class__, str], arch_params: dict, checkpoint_params: dict, *args, **kwargs) -> tuple:
        """
        Instantiates kd_module according to architecture and arch_params, handles pretrained weights for the student
         and teacher networks, and the required module manipulation (i.e head replacement) for the teacher network.

        :param architecture: String, KDModule or uninstantiated KDModule class describing the netowrks architecture.
        :param arch_params: Architecture's parameters passed to networks c'tor.
        :param checkpoint_params: checkpoint loading related parameters dictionary with 'pretrained_weights' key,
            s.t it's value is a string describing the dataset of the pretrained weights (for example "imagenent").

        :return: instantiated netowrk i.e KDModule, architecture_class (will be none when architecture is not str)
        """

        student_architecture = get_param(kwargs, "student_architecture")
        teacher_architecture = get_param(kwargs, "teacher_architecture")
        student_arch_params = get_param(kwargs, "student_arch_params")
        teacher_arch_params = get_param(kwargs, "teacher_arch_params")
        student_arch_params = core_utils.HpmStruct(**student_arch_params)
        teacher_arch_params = core_utils.HpmStruct(**teacher_arch_params)
        student_pretrained_weights = get_param(checkpoint_params, "student_pretrained_weights")
        teacher_pretrained_weights = get_param(checkpoint_params, "teacher_pretrained_weights")

        student = super()._instantiate_net(student_architecture, student_arch_params, {"pretrained_weights": student_pretrained_weights})
        teacher = super()._instantiate_net(teacher_architecture, teacher_arch_params, {"pretrained_weights": teacher_pretrained_weights})

        run_teacher_on_eval = get_param(kwargs, "run_teacher_on_eval", default_val=False)

        return self._instantiate_kd_net(arch_params, architecture, run_teacher_on_eval, student, teacher)

    def _instantiate_kd_net(self, arch_params, architecture, run_teacher_on_eval, student, teacher):
        if isinstance(architecture, str):
            architecture_cls = KD_ARCHITECTURES[architecture]
            net = architecture_cls(arch_params=arch_params, student=student, teacher=teacher, run_teacher_on_eval=run_teacher_on_eval)
        elif isinstance(architecture, KDModule.__class__):
            net = architecture(arch_params=arch_params, student=student, teacher=teacher, run_teacher_on_eval=run_teacher_on_eval)
        else:
            net = architecture
        return net

    def _load_checkpoint_to_model(self):
        """
        Initializes teacher weights with teacher_checkpoint_path if needed, then handles checkpoint loading for
         the entire KD network following the same logic as in Trainer.
        """
        teacher_checkpoint_path = get_param(self.checkpoint_params, "teacher_checkpoint_path")
        teacher_net = self.net.module.teacher

        if teacher_checkpoint_path is not None:

            #  WARN THAT TEACHER_CKPT WILL OVERRIDE TEACHER'S PRETRAINED WEIGHTS
            teacher_pretrained_weights = get_param(self.checkpoint_params, "teacher_pretrained_weights")
            if teacher_pretrained_weights:
                logger.warning(teacher_checkpoint_path + " checkpoint is " "overriding " + teacher_pretrained_weights + " for teacher model")

            # ALWAYS LOAD ITS EMA IF IT EXISTS
            load_teachers_ema = "ema_net" in read_ckpt_state_dict(teacher_checkpoint_path).keys()
            load_checkpoint_to_model(
                ckpt_local_path=teacher_checkpoint_path,
                load_backbone=False,
                net=teacher_net,
                strict="no_key_matching",
                load_weights_only=True,
                load_ema_as_net=load_teachers_ema,
            )

        super()._load_checkpoint_to_model()

    def _add_metrics_update_callback(self, phase):
        """
        Adds KDModelMetricsUpdateCallback to be fired at phase

        :param phase: Phase for the metrics callback to be fired at
        """
        self.phase_callbacks.append(KDModelMetricsUpdateCallback(phase))

    def _get_hyper_param_config(self):
        """
        Creates a training hyper param config for logging with additional KD related hyper params.
        """
        hyper_param_config = super()._get_hyper_param_config()
        hyper_param_config.update(
            {
                "student_architecture": self.student_architecture,
                "teacher_architecture": self.teacher_architecture,
                "student_arch_params": self.student_arch_params,
                "teacher_arch_params": self.teacher_arch_params,
            }
        )
        return hyper_param_config

    def _instantiate_ema_model(self, ema_params: Mapping[str, Any]) -> KDModelEMA:
        """Instantiate ema model for standard SgModule.
        :param decay_type: (str) The decay climb schedule. See EMA_DECAY_FUNCTIONS for more details.
        :param decay: The maximum decay value. As the training process advances, the decay will climb towards this value
                      according to decay_type schedule. See EMA_DECAY_FUNCTIONS for more details.
        :param kwargs: Additional parameters for the decay function. See EMA_DECAY_FUNCTIONS for more details.
        """
        logger.info(f"Using EMA with params {ema_params}")
        return KDModelEMA.from_params(self.net, **ema_params)

    def train(
        self,
        model: KDModule = None,
        training_params: Dict = None,
        student: SgModule = None,
        teacher: torch.nn.Module = None,
        kd_architecture: Union[KDModule.__class__, str] = "kd_module",
        kd_arch_params: Dict = None,
        run_teacher_on_eval=False,
        train_loader: DataLoader = None,
        valid_loader: DataLoader = None,
        additional_configs_to_log: Dict = None,
        *args,
        **kwargs,
    ):
        """
        Trains the student network (wrapped in KDModule network).


        :param model: KDModule, network to train. When none is given will initialize KDModule according to kd_architecture,
            student and teacher (default=None)
        :param training_params: dict, Same as in Trainer.train()
        :param student: SgModule - the student trainer
        :param teacher: torch.nn.Module- the teacher trainer
        :param kd_architecture: KDModule architecture to use, currently only 'kd_module' is supported (default='kd_module').
        :param kd_arch_params: architecture params to pas to kd_architecture constructor.
        :param run_teacher_on_eval: bool- whether to run self.teacher at eval mode regardless of self.train(mode)
        :param train_loader: Dataloader for train set.
        :param valid_loader: Dataloader for validation.
        :param additional_configs_to_log: Dict, dictionary containing configs that will be added to the training's
                sg_logger. Format should be {"Config_title_1": {...}, "Config_title_2":{..}}, (optional, default=None)
        """
        kd_net = self.net or model
        kd_arch_params = kd_arch_params or dict()
        if kd_net is None:
            if student is None or teacher is None:
                raise ValueError("Must pass student and teacher models or net (KDModule).")
            kd_net = self._instantiate_kd_net(
                arch_params=HpmStruct(**kd_arch_params), architecture=kd_architecture, run_teacher_on_eval=run_teacher_on_eval, student=student, teacher=teacher
            )
        super().train(
            model=kd_net,
            training_params=training_params,
            train_loader=train_loader,
            valid_loader=valid_loader,
            additional_configs_to_log=additional_configs_to_log,
        )


import logging
import os
import sys
import time
from typing import Union


def get_logger(logger_name: str, log_level: Union[str, None] = None) -> logging.Logger:
    AutoLoggerConfig.get_instance()
    logger: logging.Logger = logging.getLogger(logger_name)
    if log_level is not None:
        logger.setLevel(log_level)
    if int(os.getenv("LOCAL_RANK", -1)) > 0:
        mute_current_process()
    return logger


class AutoLoggerConfig:
    """
    A Class for the Automated Logging Config
    """

    FILE_LOGGING_LEVEL = os.environ.get("LOG_LEVEL", "DEBUG").upper()
    CONSOLE_LOGGING_LEVEL = os.environ.get("CONSOLE_LOG_LEVEL", "INFO").upper()

    filename: Union[str, None]

    def __init__(self):
        self.filename = None

    def _setup_default_logging(self, log_level: str = None) -> None:
        """
        Setup default logging configuration. Usually happens when app starts, and we don't have
        experiment dir yet.
        The default log directory will be `~/ezdl_logs`
        :param log_level: The default log level to use. If None, uses LOG_LEVEL and CONSOLE_LOG_LEVEL environment vars.
        :return: None
        """

        # There is no _easy_ way to log all events to a single file, when using DDP or DataLoader with num_workers > 1
        # on Windows platform. In both these cases a multiple processes will be spawned and multiple logs may be created.
        # Therefore the log file will have the parent PID to being able to discriminate the logs corresponding to a single run.
        timestamp = time.strftime("%Y_%m_%d_%H_%M_%S", time.localtime())
        self._setup_logging(
            filename=os.path.expanduser(f"~/ezdl_logs/ezdl_logs_{os.getppid()}_{timestamp}.log"),
            copy_already_logged_messages=False,
            filemode="w",
            log_level=log_level,
        )

    def _setup_logging(self, filename: str, copy_already_logged_messages: bool, filemode: str = "a", log_level: str = None) -> None:
        """
        Sets the logging configuration to store messages to specific file
        :param filename: Output log file
        :param filemode: Open mode for file
        :param copy_already_logged_messages: Controls whether messages from previous log configuration should be copied
               to new place. This is helpful to transfer diagnostic messages (from the app start) to experiment dir.
        :param log_level: The default log level to use. If None, uses LOG_LEVEL and CONSOLE_LOG_LEVEL environment vars.
        :return:
        """
        os.makedirs(os.path.dirname(filename), exist_ok=True)

        if copy_already_logged_messages and self.filename is not None and os.path.exists(self.filename):
            with open(self.filename, "r", encoding="utf-8") as src:
                with open(filename, "w") as dst:
                    dst.write(src.read())

        file_logging_level = log_level or self.FILE_LOGGING_LEVEL
        console_logging_level = log_level or self.CONSOLE_LOGGING_LEVEL

        cur_version = sys.version_info
        python_38 = (3, 8)
        python_39 = (3, 9)
        manager = logging.getLogger("").manager

        extra_kwargs = {}
        if cur_version >= python_38:
            extra_kwargs = dict(
                force=True,
            )
        else:
            # If the logging does not support force=True, we should manually delete handlers
            del manager.root.handlers[:]

        if cur_version >= python_39:
            extra_kwargs["encoding"] = "utf-8"

        logging.basicConfig(
            filename=filename,
            filemode=filemode,
            format="%(asctime)s %(levelname)s - %(name)s - %(message)s",
            datefmt="[%Y-%m-%d %H:%M:%S]",
            level=file_logging_level,
            **extra_kwargs,
        )

        # Add console handler
        console_handler = logging.StreamHandler(stream=sys.stdout)
        console_handler.setLevel(console_logging_level)
        console_handler.setFormatter(
            logging.Formatter(
                "%(asctime)s %(levelname)s - %(filename)s - %(message)s",
                datefmt="[%Y-%m-%d %H:%M:%S]",
            )
        )
        manager.root.handlers.append(console_handler)

        self.filename = filename

    @classmethod
    def get_instance(cls):
        global _ez_logger_config
        if _ez_logger_config is None:
            _ez_logger_config = cls()
            _ez_logger_config._setup_default_logging()

        return _ez_logger_config

    @classmethod
    def get_log_file_path(cls) -> str:
        """
        Return the current log file used to store log messages
        :return: Full path to log file
        """
        self = cls.get_instance()
        return self.filename

    @classmethod
    def setup_logging(cls, filename: str, copy_already_logged_messages: bool, filemode: str = "a", log_level: str = None) -> None:
        self = cls.get_instance()
        self._setup_logging(filename, copy_already_logged_messages, filemode, log_level)


_ez_logger_config = None


def mute_current_process():
    """Mute prints, warnings and all logs except ERRORS. This is meant when running multiple processes."""
    # Ignore warnings
    import warnings

    warnings.filterwarnings("ignore")

    # Ignore prints
    import sys

    sys.stdout = open(os.devnull, "w")

    # Only show ERRORS
    process_loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
    for logger in process_loggers:
        logger.setLevel(logging.ERROR)


from __future__ import annotations

import copy
import gc
import os
import pandas as pd

from typing import Mapping
from easydict import EasyDict

from ezdl.experiment.resume import ExpLog
from ezdl.experiment.run import Run
from ezdl.experiment.resume import get_interrupted_run, retrieve_run_to_resume
from ezdl.utils.grid import linearized_to_string, make_grid, linearize
from ezdl.utils.utilities import nested_dict_update, update_collection
from ezdl.logger.text_logger import get_logger

logger = get_logger(__name__)


class GridSummary:
    def __init__(self,
                 total_runs,
                 total_runs_excl_grid,
                 total_runs_to_run,
                 total_runs_excl,
                 ):
        self.total_runs = total_runs
        self.total_runs_excl_grid = total_runs_excl_grid
        self.total_runs_to_run = total_runs_to_run
        self.total_runs_excl = total_runs_excl

    def update(self, d):
        self.total_runs = d.get("total_runs") or self.total_runs
        self.total_runs_excl_grid = d.get("total_runs_excl_grid") or self.total_runs_excl_grid
        self.total_runs_to_run = d.get("total_runs_to_run") or self.total_runs_to_run
        self.total_runs_excl = d.get("total_runs_excl") or self.total_runs_to_run


class ExpSettings(EasyDict):
    def __init__(self, *args, **kwargs):
        self.start_from_grid = 0
        self.start_from_run = 0
        self.resume = False
        self.resume_last = False
        self.tracking_dir = ""
        self.excluded_files = ""
        self.name = ""
        self.group = ""
        self.continue_with_errors = True
        self.logger = None
        super().__init__(*args, **kwargs)
        self.tracking_dir = self.tracking_dir or ""

    def update(self, e: ExpSettings, **f):
        if e is None:
            return
        self.start_from_grid = e.start_from_grid or self.start_from_grid
        self.start_from_run = e.start_from_run or self.start_from_run
        self.resume = e.resume or self.resume
        self.resume_last = e.resume_last or self.resume_last
        self.tracking_dir = e.tracking_dir or self.tracking_dir
        self.excluded_files = e.excluded_files or self.excluded_files
        self.group = e.group or self.group
        self.logger = e.logger or self.logger
        self.continue_with_errors = not e.continue_with_errors or self.continue_with_errors


class Status:
    STARTING = "starting"
    CRASHED = "crashed"
    FINISHED = "finished"

    def __init__(self, grid, run, params, n_grids, grid_len, run_name=None, run_url=None):
        self.grid = grid
        self.run = run
        self.params = params
        self.status = self.STARTING
        self.grid_len = grid_len
        self.n_grids = n_grids
        self.exception = None
        self.run_name = run_name
        self.run_url = run_url

    def finish(self):
        self.params = {}
        self.status = self.FINISHED
        return self

    def crash(self, exception):
        self.status = self.CRASHED
        self.exception = exception
        return self


class StatusManager:
    def __init__(self, n_grids):
        self.n_grids = n_grids
        self.cur_status = None

    def new_run(self, grid, run, params, grid_len, run_name=None, run_url=None):
        self.cur_status = Status(grid=grid, run=run, params=params,
                                 n_grids=self.n_grids, grid_len=grid_len,
                                 run_name=run_name, run_url=run_url)
        return self.cur_status

    def update_run(self, run_name, run_url):
        self.cur_status.run_name = run_name
        self.cur_status.run_url = run_url
        return self.cur_status

    def finish_run(self):
        return self.cur_status.finish()

    def crash_run(self, exception):
        return self.cur_status.crash(exception)


class Experimenter:
    EXP_FINISH_SEP = "#"*50 + " FINISHED " + "#"*50 + "\n"
    EXP_CRASHED_SEP = "|\\"*50 + "CRASHED" + "|\\"*50 + "\n"

    def __init__(self):
        self.gs = None
        self.exp_settings = ExpSettings()
        self.grids = None

    def calculate_runs(self, settings):
        base_grid = settings['parameters']
        other_grids = settings['other_grids']
        self.exp_settings = ExpSettings(settings['experiment'])

        print('\n' + '='*100)
        complete_grids = [base_grid]
        if other_grids:
            complete_grids += \
                [nested_dict_update(copy.deepcopy(base_grid), other_run) for other_run in other_grids]
        logger.info(f'There are {len(complete_grids)} grids')

        self.grids, dot_elements = zip(*[make_grid(grid, return_cartesian_elements=True) for grid in complete_grids])
        # WARNING: Grids' objects have the same IDs!
        dot_elements = list(dot_elements)
        if len(dot_elements) > 1:
            dot_elements[1:] = [list(dict(linearize(others) + dot).items()) for others, dot in
                                zip(other_grids, dot_elements[1:])]

        # Modify starting grid and run to manage the resume
        self.manage_resume()

        for i, grid in enumerate(self.grids):
            info = f'Found {len(grid)} runs from grid {i}'
            last_grid = self.exp_settings.start_from_grid if self.exp_settings.start_from_grid is not None else len(self.grids)
            if i < last_grid:
                info += f', skipping grid {i} with {len(grid)} runs'
            logger.info(info)
        self.generate_grid_summary()

        # logger.info(f'Total runs found:              {self.gs.total_runs}')
        # logger.info(f'Total runs excluded by grids:  {self.gs.total_runs_excl_grid}')
        # logger.info(f'Total runs excluded:           {self.gs.total_runs_excl}')
        # logger.info(f'Total runs to run:             {self.gs.total_runs_to_run}')

        if self.exp_settings.excluded_files:
            os.environ['WANDB_IGNORE_GLOBS'] = self.exp_settings.excluded_files
            
        print_preview(self, self.gs, self.grids, dot_elements)
        print('='*100 + '\n')
        
        return self.gs, self.grids, dot_elements

    def generate_grid_summary(self):
        total_runs = sum(len(grid) for grid in self.grids)
        if self.exp_settings.start_from_grid is None:
            total_runs_excl_grid = total_runs - len(self.grids[-1])
            total_runs_excl = total_runs
        else:
            total_runs_excl_grid = total_runs - sum(
                len(grid) for grid in self.grids[self.exp_settings.start_from_grid :]
            )
            total_runs_excl = total_runs_excl_grid + self.exp_settings.start_from_run
        total_runs_to_run = total_runs - total_runs_excl
        self.gs = GridSummary(
            total_runs=total_runs,
            total_runs_excl_grid=total_runs_excl_grid,
            total_runs_to_run=total_runs_to_run,
            total_runs_excl=total_runs_excl
        )

    def execute_runs_generator(self):
        track_dir = self.exp_settings['tracking_dir']
        if track_dir:
            os.makedirs(track_dir, exist_ok=True)
        exp_log = ExpLog(track_dir, self.exp_settings.name, self.exp_settings.group)
        starting_run = self.exp_settings.start_from_run
        status_manager = StatusManager(len(self.grids))
        if self.exp_settings.resume_last:
            logger.info("+ another run to finish!")
            grid_list = [(i, j) for i in range(len(self.grids)) for j in range(len(self.grids[i]))]
            if self.exp_settings.start_from_grid is None:
                grid_len = len(self.grids[-1])
                sg, sr = grid_list[-1]
            else:
                grid_len = len(self.grids[self.exp_settings.start_from_grid])
                index = grid_list.index((self.exp_settings.start_from_grid, self.exp_settings.start_from_run))
                sg, sr = grid_list[index - 1]
            try:
                exp_log.insert_run(sg, sr)
                run = get_interrupted_run(self.exp_settings)
                yield status_manager.new_run(sg, sr, run.params, grid_len, run.seg_trainer.sg_logger.name, run.seg_trainer.sg_logger.url)
                logger.info(f'Running grid {sg} out of {len(self.grids) - 1}')
                logger.info(
                    f'Running run {sr - 1} out of {grid_len} ({sum(len(self.grids[k]) for k in range(sg)) + sr} / {self.gs.total_runs - 1})'
                )
                run.launch()
                print(self.EXP_FINISH_SEP)
                exp_log.finish_run(sg, sr)
                yield status_manager.finish_run()
            except Exception as ex:
                logger.error(f'Experiment {sg} failed with error {ex}')
                print(self.EXP_CRASHED_SEP)
                exp_log.finish_run(sg, sr, crashed=True)
                if not self.exp_settings.continue_with_errors:
                    raise ex
                yield status_manager.crash_run(ex)
        for i in range(self.exp_settings.start_from_grid, len(self.grids)):
            grid = self.grids[i]
            if i != self.exp_settings.start_from_grid:
                starting_run = 0
            for j in range(starting_run, len(grid)):
                params = grid[j]
                try:
                    exp_log.insert_run(i, j)
                    yield status_manager.new_run(i, j, params, len(grid))
                    logger.info(f'Running grid {i} out of {len(self.grids) - 1}')
                    logger.info(
                        f'Running run {j} out of {len(grid) - 1} ({sum(len(self.grids[k]) for k in range(i)) + j} / {self.gs.total_runs - 1})'
                    )
                    run = Run()
                    run.init({'experiment': {**self.exp_settings}, **params})
                    yield status_manager.update_run(run.seg_trainer.sg_logger.name, run.seg_trainer.sg_logger.url)
                    run.launch()
                    print(self.EXP_FINISH_SEP)
                    exp_log.finish_run(i, j)
                    gc.collect()
                    yield status_manager.finish_run()
                except Exception as ex:
                    logger.error(f'Experiment {i} failed with error {ex}')
                    print(self.EXP_CRASHED_SEP)
                    exp_log.finish_run(i, j, crashed=True)
                    if not self.exp_settings.continue_with_errors:
                        raise ex
                    yield status_manager.crash_run(ex)

    def execute_runs(self):
        for _ in self.execute_runs_generator():
            pass

    def manage_resume(self):
        if self.exp_settings.resume:
            self.exp_settings.start_from_grid, \
                self.exp_settings.start_from_run, \
                self.exp_settings.resume_last = retrieve_run_to_resume(self.exp_settings, self.grids)
        else:
            self.exp_settings.resume_last = False

    def update_settings(self, d):
        self.exp_settings = update_collection(self.exp_settings, d)
        if self.gs is None:
            return
        self.gs.update(self.exp_settings)
        if "resume" in d:
            self.manage_resume()
            self.generate_grid_summary()


def experiment(settings: Mapping, param_path: str = "local variable"):
    logger.info(f'Loaded parameters from {param_path}')

    experimenter = Experimenter()
    grid_summary, grids, cartesian_elements = experimenter.calculate_runs(settings)

    experimenter.execute_runs()


def preview(settings: Mapping, param_path: str = "local variable"):
    print(f'Loaded parameters from {param_path}')

    experimenter = Experimenter()
    _, _, _ = experimenter.calculate_runs(settings)
    
    
def print_preview(experimenter, grid_summary, grids, cartesian_elements):
    summary_series = pd.concat([pd.Series(grid_summary), pd.Series(experimenter.exp_settings.__dict__)])
    summary_string = f"\n{summary_series.to_string()}\n"
    
    dfs = [pd.DataFrame(linearized_to_string(dot_element), columns=[f"Grid {i}", f"N. runs: {len(grid)}"])
           for i, (dot_element, grid) in enumerate(zip(cartesian_elements, grids))]
    mark_grids = "\n\n".join(df.to_string(index=False) for df in dfs)
    mark_grids = "Most important parameters for each grid \n" + mark_grids
    logger.info(f"\n{summary_string}\n{mark_grids}")



from super_gradients.training.models.kd_modules.kd_module import KDModule
from super_gradients.training.utils import HpmStruct


class LogitsDistillationModule(KDModule):
    """
    Logits Distillation Module
    """
    def __init__(self, arch_params, student, teacher, **kwargs):
        super().__init__(arch_params, student, teacher, **kwargs)
        if hasattr(self.student.module, "update_param_groups"):
            self.update_param_groups = self.student_update_param_groups
        else:
            self.update_param_groups = self.default_update_param_groups

    def initialize_param_groups(self, lr: float, training_params: HpmStruct) -> list:
        if hasattr(self.student.module, "initialize_param_groups"):
            return self.student.module.initialize_param_groups(lr, training_params)
        else:
            return [{"named_params": self.student.named_parameters()}]

    def default_update_param_groups(self, param_groups: list, lr: float, epoch: int, iter: int, training_params: HpmStruct,
                            total_batch: int) -> list:  
        for param_group in param_groups:
            param_group["lr"] = lr
        return param_groups

    def student_update_param_groups(self, param_groups: list, lr: float, epoch: int, iter: int, training_params: HpmStruct,
                            total_batch: int) -> list:
        return self.student.update_param_groups(param_groups, lr, epoch, iter, training_params, total_batch)
    


import numpy as np
import torchvision.models as models
import inspect
import torch

from inspect import signature
from ptflops import get_model_complexity_info
from ezdl.models import MODELS
from ezdl.utils.utilities import load_yaml


def seg_model_flops(model, size, verbose=False, per_layer_stats=False):
    n_channels, w, h = size
    macs, params = get_model_complexity_info(model, (n_channels, w, h), as_strings=True,
                                             print_per_layer_stat=per_layer_stats, verbose=verbose)
    print('{:<30}  {:<8}'.format('Computational complexity: ', macs))
    print('{:<30}  {:<8}'.format('Number of parameters: ', params))


def model_parameters(model): 
    total_params = sum(
        param.numel() for param in model.parameters()
    )
    print(f'{total_params:,} total parameters.')


def seg_inference_throughput(model, size, batch_size, device):
    n_channels, w, h = size
    dummy_input = torch.randn(batch_size, n_channels, w, h, dtype=torch.float).to(device)
    repetitions = 100
    warmup = 50
    total_time = 0
    with torch.no_grad():
        for rep in range(warmup):
            _ = model(dummy_input)
        for rep in range(repetitions):
            starter, ender = torch.cuda.Event(enable_timing=True), torch.cuda.Event(enable_timing=True)
            starter.record()
            _ = model(dummy_input)
            ender.record()
            torch.cuda.synchronize()
            curr_time = starter.elapsed_time(ender) / 1000
            total_time += curr_time
    throughput = (repetitions * batch_size) / total_time
    print('Final Throughput:', throughput)


def seg_inference_inference_per_second(model, size, batch_size, device, model_args={}):
    n_channels, w, h = size
    dummy_input = torch.randn(batch_size, n_channels, w, h, dtype=torch.float).to(device)
    starter, ender = torch.cuda.Event(enable_timing=True), torch.cuda.Event(enable_timing=True)
    repetitions = 500
    timings = np.zeros((repetitions, 1))
    # GPU-WARM-UP
    for _ in range(50):
        _ = model(dummy_input)
    # MEASURE PERFORMANCE
    with torch.no_grad():
        for rep in range(repetitions):
            starter.record()
            _ = model(dummy_input)
            ender.record()
            # WAIT FOR GPU SYNC
            torch.cuda.synchronize()
            curr_time = starter.elapsed_time(ender)
            timings[rep] = curr_time
    mean_syn = np.sum(timings) / repetitions
    std_syn = np.std(timings)
    print(f"Mean inference time: {mean_syn} ms")
    print(f"Time per example {mean_syn / batch_size} ms")


def complexity(file):
    params = load_yaml(file)
    models = [
        # ('lawin', 1, {'backbone_pretrained': True}),
        # ('lawin', 2, {'backbone_pretrained': True}),
        # ('lawin', 3, {'backbone_pretrained': True}),
        # ('lawin', 4, {'backbone_pretrained': True, 'main_pretrained': ['R', 'G', 'G', 'G']}),
        #
        # ('lawin', 1, {'backbone': 'MiT-B1'}),
        # ('lawin', 2, {'backbone': 'MiT-B1'}),
        # ('lawin', 3, {'backbone': 'MiT-B1'}),
        # ('lawin', 4, {'backbone': 'MiT-B1'}),
        #
        # ('laweed', 1, {'backbone_pretrained': True}),
        # ('laweed', 2, {'backbone_pretrained': True}),
        # ('laweed', 3, {'backbone_pretrained': True}),
        # ('laweed', 4, {'backbone_pretrained': True, 'main_pretrained': ['R', 'G', 'G', 'G']}),
        #
        # ('laweed', 1, {'backbone_pretrained': True, 'backbone': 'MiT-B1'}),
        # ('laweed', 2, {'backbone_pretrained': True, 'backbone': 'MiT-B1'}),
        # ('laweed', 3, {'backbone_pretrained': True, 'backbone': 'MiT-B1'}),
        # ('laweed', 4, {'backbone_pretrained': True, 'main_pretrained': ['R', 'G', 'G', 'G'], 'backbone': 'MiT-B1'}),
        #
        #
        # ('splitlawin', 3, {'main_channels': 2}),
        # ('splitlawin', 4, {'main_channels': 2}),
        # ('splitlawin', 3, {'main_channels': 2, 'backbone': 'MiT-B1'}),
        # ('splitlawin', 4, {'main_channels': 2, 'backbone': 'MiT-B1'}),
        #
        # ('splitlaweed', 3, {'main_channels': 2, 'main_pretrained': ['R', 'G'], 'side_pretrained': 'G'}),
        # ('splitlaweed', 4, {'main_channels': 2, 'main_pretrained': ['R', 'G'], 'side_pretrained': 'G'}),
        #
        # ('doublelawin', 3, {'main_channels': 2}),
        # ('doublelawin', 4, {'main_channels': 2}),
        # ('doublelawin', 3, {'main_channels': 2, 'backbone': 'MiT-B1'}),
        # ('doublelawin', 4, {'main_channels': 2, 'backbone': 'MiT-B1'}),
        #
        # ('doublelaweed', 3, {'main_channels': 2, 'main_pretrained': ['R', 'G'], 'side_pretrained': 'G'}),
        # ('doublelaweed', 4, {'main_channels': 2, 'main_pretrained': ['R', 'G'], 'side_pretrained': 'G'}),
        # ('segnet', 1, {}),
        # ('segnet', 2, {}),
        # ('segnet', 3, {}),
        # ('segnet', 4, {}),
        # ('resnet', 3, {"model_name": "50"}),
        # ('deeplabv3_resnet50', 3, {}, 6),
        # ('splitlawin', 5, {'main_channels': 3, 'backbone': 'MiT-B0'}),
        # ('splitlawin', 5, {'main_channels': 3, 'backbone': 'MiT-B1'}),
        # ('lawin', 5, {'backbone': 'MiT-B0'}, 16),
        # ('lawin', 5, {'backbone': 'MiT-B1'}, 8),
        # ('lawin', 5, {'backbone': 'MiT-LD'}, 32),
        # ('lawin', 5, {'backbone': 'MiT-L0'}, 48),
        # ('lawin', 5, {'backbone': 'MiT-L1'}, 64),
        # ('mit', 5, {'model_name': 'B0'}),
        # ('mit', 5, {'model_name': 'B1'}),
        # ('mit', 5, {'model_name': 'LD'}),
        # ('mit', 5, {'model_name': 'LT'}),
        # ('mit', 5, {'model_name': 'L0'}),
        # ('mit', 5, {'model_name': 'L1'}),
    ]
    per_layer_stats = params.get('per_layer_stats', False)
    verbose = params.get('verbose', False)
    models = params['models']
    wh = params.get("size") or [256, 256]
    default_batch_size = params.get('batch_size', 1)
    defualt_channels = params.get('in_channels', 3)
    default_num_classes = params.get('num_classes', 3)

    for params in models:
        model = params['name']
        batch_size = params.get('batch_size', default_batch_size)
        in_channels = params.get('in_channels', defualt_channels)
        num_classes = params.get('num_classes', default_num_classes)
        params = params.get('params') or {}

        with torch.cuda.device(0):
            if inspect.isclass(MODELS[model]):
                model_signature = signature(MODELS[model].__init__).parameters.keys()
            else:
                model_signature = signature(MODELS[model]).parameters.keys()
            in_params = {'input_channels': in_channels, 'num_classes': num_classes, "output_channels": num_classes, **params}
            if 'arch_params' in model_signature:
                net = MODELS[model](in_params).to('cuda')
            else:
                actual_args = {k: v for k, v in in_params.items() if k in model_signature}
                discard_args = {k: v for k, v in in_params.items() if k not in model_signature}
                print(f"Discarded args: {discard_args}")
                net = MODELS[model](**actual_args).to('cuda')
            size = [in_channels] + wh
            print(f"Model: {model}")
            print(f"args: {params}")
            print(f"Batches: {batch_size}")
            print(f"Size: {size}")
            seg_model_flops(net, size, verbose, per_layer_stats)
            model_parameters(net)
            # seg_inference_throughput(model, channels, batch_size, 'cuda', args)
            seg_inference_inference_per_second(net, size, batch_size, 'cuda')
            torch.cuda.empty_cache()
            print()
        del model


if __name__ == '__main__':
    file = "complexity.yaml"
    complexity(file)


from ezdl.models.kd.feature import FDOutput
from ezdl.utils.utilities import substitute_values, instantiate_class
from einops import rearrange

from torch.nn import CrossEntropyLoss, Module, MSELoss
import torch.nn.functional as F
import torch.nn as nn
import torch

from ezdl.logger.text_logger import get_logger


logger = get_logger(__name__)


def get_reduction(reduction: str):
    if reduction == 'none':
        return lambda x: x
    elif reduction == 'mean':
        return torch.mean
    elif reduction == 'sum':
        return torch.sum
    else:
        raise NotImplementedError(f"Invalid reduction mode: {reduction}")


class CELoss(CrossEntropyLoss):
    def __init__(self, *args, weight=None, **kwargs):
        if weight:
            weight = torch.tensor(weight)
        super().__init__(*args, weight=weight, **kwargs)


class FocalLoss(Module):
    def __init__(self, gamma: float = 0, weight=None, reduction: str = 'mean', **kwargs):
        super().__init__()
        self.weight = None
        if weight:
            self.weight = torch.tensor(weight)
        self.gamma = gamma

        self.reduction = get_reduction(reduction)

    def __call__(self, x, target, **kwargs):
        ce_loss = F.cross_entropy(x, target, reduction='none', **kwargs)
        pt = torch.exp(-ce_loss)
        if self.weight is not None:
            self.weight = self.weight.to(x.device)
            wtarget = substitute_values(target, self.weight, unique=torch.arange(len(self.weight), device=target.device))
            focal_loss = torch.pow((1 - pt), self.gamma) * wtarget * ce_loss
        else:
            focal_loss = torch.pow((1 - pt), self.gamma) * ce_loss

        return self.reduction(focal_loss)


# based on:
# https://github.com/kevinzakka/pytorch-goodies/blob/master/losses.py
class DiceLoss(nn.Module):
    r"""Criterion that computes Sørensen-Dice Coefficient loss.

    According to [1], we compute the Sørensen-Dice Coefficient as follows:

    .. math::

        \text{Dice}(x, class) = \frac{2 |X| \cap |Y|}{|X| + |Y|}

    where:
       - :math:`X` expects to be the scores of each class.
       - :math:`Y` expects to be the one-hot tensor with the class labels.

    the loss, is finally computed as:

    .. math::

        \text{loss}(x, class) = 1 - \text{Dice}(x, class)

    [1] https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient

    Shape:
        - Input: :math:`(N, C, H, W)` where C = number of classes.
        - Target: :math:`(N, H, W)` where each value is
          :math:`0 ≤ targets[i] ≤ C−1`.

    Examples:
        >>> N = 5  # num_classes
        >>> loss = tgm.losses.DiceLoss()
        >>> input = torch.randn(1, N, 3, 5, requires_grad=True)
        >>> target = torch.empty(1, 3, 5, dtype=torch.long).random_(N)
        >>> output = loss(input, target)
        >>> output.backward()
    """

    def __init__(self, weight = None, reduction: str = 'mean', average: str = 'micro') -> None:
        super(DiceLoss, self).__init__()
        self.weight = None
        self.average = average
        if weight:
            self.weight = torch.tensor(weight)
            if self.average == 'micro':
                raise ValueError("Weighted Dice Loss is only supported for macro average")
        self.reduction = get_reduction(reduction)
        self.eps: float = 1e-6

    def forward(
            self,
            input: torch.Tensor,
            target: torch.Tensor) -> torch.Tensor:
        if not torch.is_tensor(input):
            raise TypeError("Input type is not a torch.Tensor. Got {}"
                            .format(type(input)))
        if not len(input.shape) == 4:
            raise ValueError("Invalid input shape, we expect BxNxHxW. Got: {}"
                             .format(input.shape))
        if not input.shape[-2:] == target.shape[-2:]:
            raise ValueError("input and target shapes must be the same. Got: {}"
                             .format(input.shape, input.shape))
        if not input.device == target.device:
            raise ValueError(
                "input and target must be in the same device. Got: {}" .format(
                    input.device, target.device))
        # compute softmax over the classes axis
        input_soft = F.softmax(input, dim=1)

        # create the labels one hot tensor
        target_one_hot = F.one_hot(target, num_classes=input.shape[1]).permute(0, 3, 2, 1)

        if self.average == 'macro':
            return self._macro_forward(input_soft, target_one_hot)
        # compute the actual dice score
        dice_score = self._calc_dice(input_soft, target_one_hot)
        return self.reduction(1. - dice_score)

    def _calc_dice(self, input, target):
        dims = (1, 2, 3)

        intersection = torch.sum(input * target, dims)
        cardinality = torch.sum(input + target, dims)
        dice_score = (2. * intersection + self.eps) / (cardinality + self.eps)
        return dice_score
    
    def _macro_forward(self, input, target):
        flat_input = rearrange(input, 'b (c bin) h w -> (b c) bin h w', bin=1)
        flat_target = rearrange(target, 'b (c bin) h w -> (b c) bin h w', bin=1)

        dice = 1. - self._calc_dice(flat_input, flat_target) # (B X C)
        dice = rearrange(dice, '(b c) -> b c', c=input.shape[1])
        if self.weight is not None:
            dice = dice * self.weight.to(input.device)
        dice = dice.mean(dim=1)
        return self.reduction(dice)



class VisklDivLoss(nn.KLDivLoss):
    """ KL divergence wrapper for Computer Vision tasks."""
    def __init__(self, reduction: str = 'mean', **kwargs):
        super().__init__(reduction='none')
        self.macro_reduction = get_reduction(reduction)

    def forward(self, student_output, teacher_output):
        return self.macro_reduction(
            super().forward(torch.log_softmax(student_output, dim=1),
                                                torch.softmax(teacher_output, dim=1)).sum(dim=1)
        )
        

class ComposedLoss(nn.Module):
    name = "ComposedLoss"
    def __init__(self) -> None:
        super().__init__()
        self.__class__.__name__ = self.name
        
    def component_names(self):
        raise NotImplementedError("Component names not implemented for ComposedLoss abstract class")


class AuxiliaryLoss(ComposedLoss):
    name = "AuxLoss"
    """ Auxiliary loss, wraps the task loss and auxiliary loss """
    def __init__(self, task_loss_fn, aux_loss_fn, aux_loss_coeff: float = 0.2, **kwargs):
        super().__init__()
        self.task_loss_fn = task_loss_fn
        self.aux_loss_fn = aux_loss_fn
        self.aux_loss_coeff = aux_loss_coeff

    @property
    def component_names(self):
        """
        Component names for logging during training.
        These correspond to 2nd item in the tuple returned in self.forward(...).
        See super_gradients.Trainer.train() docs for more info.
        """
        return [self.name,
        self.task_loss_fn.__class__.__name__,
        self.aux_loss_fn.__class__.__name__
        ]   

    def forward(self, task_aux_output, target):
        out, aux = task_aux_output
        task_loss = self.task_loss_fn(out, target)
        if isinstance(task_loss, tuple):  # SOME LOSS FUNCTIONS RETURNS LOSS AND LOG_ITEMS
            task_loss = task_loss[0]
        aux_loss = self.aux_loss_fn(aux['aux'], target)
        loss = task_loss * (1 - self.aux_loss_coeff) + aux_loss * self.aux_loss_coeff

        return loss, torch.cat((loss.unsqueeze(0), task_loss.unsqueeze(0), aux_loss.unsqueeze(0))).detach()


class KDLogitsLoss(ComposedLoss):
    name = "KDLLoss"
    """ Knowledge distillation loss, wraps the task loss and distillation loss """
    def __init__(self, task_loss_fn, distillation_loss_fn=VisklDivLoss(), distillation_loss_coeff: float = 0.5):
        '''
        :param task_loss_fn: task loss. E.g., LabelSmoothingCrossEntropyLoss
        :param distillation_loss_fn: distillation loss. E.g., KLDivLoss
        :param distillation_loss_coeff:
        '''

        super(KDLogitsLoss, self).__init__()
        self.task_loss_fn = task_loss_fn
        self.distillation_loss_fn = distillation_loss_fn
        self.distillation_loss_coeff = distillation_loss_coeff

    @property
    def component_names(self):
        """
        Component names for logging during training.
        These correspond to 2nd item in the tuple returned in self.forward(...).
        See super_gradients.Trainer.train() docs for more info.
        """
        return [self.name,
        self.task_loss_fn.__class__.__name__,
        self.distillation_loss_fn.__class__.__name__
        ]

    def forward(self, kd_module_output, target):
        task_loss = self.task_loss_fn(kd_module_output.student_output, target)
        if isinstance(task_loss, tuple):  # SOME LOSS FUNCTIONS RETURNS LOSS AND LOG_ITEMS
            task_loss = task_loss[0]
        distillation_loss = self.distillation_loss_fn(kd_module_output.student_output, kd_module_output.teacher_output)
        loss = task_loss * (1 - self.distillation_loss_coeff) + distillation_loss * self.distillation_loss_coeff

        return loss, torch.cat((loss.unsqueeze(0), task_loss.unsqueeze(0), distillation_loss.unsqueeze(0))).detach()


class KDFeatureLogitsLoss(ComposedLoss):
    name = "KDFLLoss"
    def __init__(self, task_loss_fn, feature_loss_fn=MSELoss(), distillation_loss_fn=VisklDivLoss(),
                 dist_feats_loss_coeffs=(0.2, 0.4, 0.4), feats_loss_reduction='mean', **kwargs):
        super().__init__()
        self.task_loss_fn = task_loss_fn
        self.feature_loss_fn = feature_loss_fn
        self.distillation_loss_fn = distillation_loss_fn
        self.dist_feats_loss_coeff = dist_feats_loss_coeffs
        self.feats_loss_reduction = get_reduction(feats_loss_reduction)

    @property
    def component_names(self):
        """
        Component names for logging during training.
        These correspond to 2nd item in the tuple returned in self.forward(...).
        See super_gradients.Trainer.train() docs for more info.
        """
        return [self.name, 
        self.task_loss_fn.__class__.__name__, 
        self.distillation_loss_fn.__class__.__name__, 
        self.feature_loss_fn.__class__.__name__
        ]

    def forward(self, kd_output: FDOutput, target):
        logits_loss = self.distillation_loss_fn(kd_output.student_output, kd_output.teacher_output)
        feats_loss = self.feats_loss_reduction(torch.tensor([
            self.feature_loss_fn(student_feat, teacher_feat)
            for student_feat, teacher_feat in zip(kd_output.student_features, kd_output.teacher_features
                                                  )], device=kd_output.student_output.device))
        task_loss = self.task_loss_fn(kd_output.student_output, target)

        loss = task_loss * self.dist_feats_loss_coeff[0] + \
               logits_loss * self.dist_feats_loss_coeff[1] + \
               feats_loss * self.dist_feats_loss_coeff[2]

        return loss, torch.cat((loss.unsqueeze(0),
                                task_loss.unsqueeze(0),
                                logits_loss.unsqueeze(0),
                                feats_loss.unsqueeze(0)
                                )).detach()


class VariationalInformationLoss(nn.Module):
    def forward(self, student_feats, teacher_feats, **kwargs):
        mu, sigma = student_feats
        return torch.sum(
            torch.log(sigma) + torch.div(torch.square(teacher_feats - mu), 2 * torch.square(sigma))
        )


class VariationalInformationLossMean(nn.Module):
    def forward(self, student_feats, teacher_feats, **kwargs):
        mu, sigma = student_feats
        return torch.mean(
            torch.log(sigma) + torch.div(torch.square(teacher_feats - mu), 2 * torch.square(sigma))
        )


class VariationalInformationLossScaled(VariationalInformationLoss):
    def __init__(self, scale_factor=1e-6, *args, **krwargs) -> None:
        super().__init__(*args, **krwargs)
        self.factor = scale_factor

    def forward(self, student_feats, teacher_feats, **kwargs):
        return super().forward(student_feats, teacher_feats, **kwargs) * self.factor


class VariationalInformationLogitsLoss(KDFeatureLogitsLoss):
    name = "VILLoss"
    variants = {
        'standard': VariationalInformationLoss,
        'mean': VariationalInformationLossMean,
        'scaled': VariationalInformationLossScaled,
    }
    def __init__(self, task_loss_fn, distillation_loss_fn=VisklDivLoss(),
                 dist_feats_loss_coeffs=(0.2, 0.4, 0.4), feats_loss_reduction='mean',
                 variant='standard', scale_factor=1e-6):
        params = {'scale_factor': scale_factor} if variant == 'scaled' else {}
        feature_loss_fn = self.variants[variant](**params)
        super().__init__(task_loss_fn=task_loss_fn,
                         feature_loss_fn=feature_loss_fn,
                         distillation_loss_fn=distillation_loss_fn,
                         dist_feats_loss_coeffs=dist_feats_loss_coeffs,
                         feats_loss_reduction=feats_loss_reduction)

    @property
    def component_names(self):
        """
        Component names for logging during training.
        These correspond to 2nd item in the tuple returned in self.forward(...).
        See super_gradients.Trainer.train() docs for more info.
        """
        comps = super().component_names
        comps[0] = self.name
        return comps


LOSSES = {
    'cross_entropy': CELoss,
    'dice': DiceLoss,
    'focal': FocalLoss,
    'variational_information_loss': VariationalInformationLoss,
    'mean_variational_information_loss': VariationalInformationLossMean,
    'scaled_variational_information_loss': VariationalInformationLossScaled,
    "vis_kldiv_loss": VisklDivLoss,

    # KD composed losses
    'kd_logits_loss': KDLogitsLoss,
    'kd_feature_logits_loss': KDFeatureLogitsLoss,
    'variational_information_logits_loss': VariationalInformationLogitsLoss,
    
    # Auxiliary losses
    "auxiliary_loss": AuxiliaryLoss
}


def instiantiate_loss(loss_name, params):
    """
    Instantiate a loss function from a string name.
    Args:
        loss_name (str): Name of the loss function.
        params (dict): Parameters for the loss function.
    Returns:
        loss_fn (nn.Module): Loss function.
    """
    try:
        return instantiate_class(loss_name, params)
    except (AttributeError, ValueError) as ex:
        logger.info(f'Loss {loss_name} not instantiable from local module.')

        loss_cls_names = {loss_cls.__name__: loss_cls for loss_cls in LOSSES.values()}
        if loss_name in LOSSES:
            return LOSSES[loss_name](**params)
        elif loss_name in loss_cls_names:
            return loss_cls_names[loss_name](**params)
        elif loss_name in nn.__dict__:
            return nn.__dict__[loss_name](**params)
        else:
            raise ValueError(
                f'Loss {loss_name} not found. Available losses: {list(LOSSES.keys())}'
            ) from ex



import torch.utils.data as torch_data

from super_gradients.training import dataloaders

from ezdl.data import DatasetInterface


class Cifar10(DatasetInterface):
    def __init__(self, dataset_params):
        super().__init__(dataset_params)
        self.train_loader = dataloaders.get("cifar10_train", dataset_params=dataset_params['trainset'],
                                            dataloader_params=dataset_params['trainloader'])
        self.test_loader = dataloaders.get("cifar10_val", dataset_params=dataset_params['testset'],
                                           dataloader_params=dataset_params['testloader'])
        self.val_loader = dataloaders.get("cifar10_val", dataset_params=dataset_params['testset'],
                                           dataloader_params=dataset_params['testloader'])
        if dataset_params.get('reduce'):
            self.train_loader = torch_data.DataLoader(
                torch_data.Subset(self.train_loader.dataset, range(500)),
                **dataset_params['trainloader']
            )
            self.test_loader = torch_data.DataLoader(
                torch_data.Subset(self.test_loader.dataset, range(20)),
                **dataset_params['testloader']
            )
            self.val_loader = torch_data.DataLoader(
                torch_data.Subset(self.val_loader.dataset, range(20)),
                **dataset_params['testloader']
            )        
    @property
    def size(self):
        return (3, 32, 32)


import importlib
from typing import Mapping, Dict, Tuple

import pandas as pd
from super_gradients.common import MultiGPUMode
from torch.cuda.amp import autocast

from piptools.scripts.sync import _get_installed_distributions

from super_gradients.common.sg_loggers import SG_LOGGERS, BaseSGLogger
from super_gradients.common.sg_loggers.abstract_sg_logger import AbstractSGLogger
from super_gradients.training import StrictLoad, Trainer

from super_gradients.training.params import TrainingParams
from super_gradients.training.utils.callbacks import Phase, PhaseContext, CallbackHandler
from super_gradients.training import utils as core_utils
from super_gradients.training.utils import sg_trainer_utils
from super_gradients.common.environment.device_utils import device_config
from super_gradients.training import models
from super_gradients.training.utils.distributed_training_utils import setup_device
from super_gradients.common.abstractions.abstract_logger import get_logger
from super_gradients.common.data_types.enum import MultiGPUMode, StrictLoad, EvaluationType
from super_gradients.common.environment.device_utils import device_config
from super_gradients.common.sg_loggers import SG_LOGGERS
from super_gradients.common.sg_loggers.abstract_sg_logger import AbstractSGLogger
from super_gradients.common.sg_loggers.base_sg_logger import BaseSGLogger
from super_gradients.training import utils as core_utils, models
from super_gradients.training.metrics.metric_utils import (
    get_metrics_results_tuple,
    get_logging_values,
    get_train_loop_description_dict,
)
from super_gradients.training.params import TrainingParams
from super_gradients.training.utils import sg_trainer_utils
from super_gradients.training.utils.callbacks import (
    CallbackHandler,
    Phase,
    PhaseContext,
)
from super_gradients.training.utils.checkpoint_utils import (
    load_checkpoint_to_model,
)
from super_gradients.training.utils.distributed_training_utils import (
    reduce_results_tuple_for_ddp,
    setup_device,
    get_gpu_mem_utilization,
    get_device_ids,
)


import torch
from super_gradients.training.utils.checkpoint_utils import load_checkpoint_to_model
from torchmetrics import MetricCollection
from tqdm import tqdm
from pprint import pformat

# from ezdl.callbacks import SegmentationVisualizationCallback
from ezdl.models import MODELS as MODELS_DICT, WrappedModel
from ezdl.callbacks import AuxMetricsUpdateCallback, callback_factory
from ezdl.utils.utilities import get_module_class_from_path, instantiate_class
from ezdl.logger.basesg_logger import BaseSGLogger as BaseLogger
from ezdl.logger import LOGGERS
from ezdl.logger.text_logger import get_logger

logger = get_logger(__name__)


class SegmentationTrainer(Trainer):
    def __init__(self, ckpt_root_dir=None, model_checkpoints_location: str = 'local', num_devices=1, multi_gpu=MultiGPUMode.AUTO, **kwargs):
        self.run_id = None
        self.num_devices = num_devices
        self.multi_gpu = multi_gpu
        setup_device(
            num_gpus=num_devices,
            multi_gpu=multi_gpu
        )
        self.train_initialized = False
        self.validation = False
        self.model_checkpoints_location = model_checkpoints_location
        super().__init__(ckpt_root_dir=ckpt_root_dir, **kwargs)

    def init_dataset(self, dataset_interface, dataset_params):
        if isinstance(dataset_interface, str):
            try:
                module, dataset_interface_cls = get_module_class_from_path(dataset_interface)
                dataset_module = importlib.import_module(module)
                dataset_interface_cls = getattr(dataset_module, dataset_interface_cls)
            except AttributeError as e:
                raise AttributeError("No interface found!") from e
            dataset_interface = dataset_interface_cls(dataset_params)
        elif isinstance(dataset_interface, type):
            pass
        else:
            raise ValueError("Dataset interface should be str or class")
        data_loader_num_workers = dataset_params.get('num_workers') or 0
        self.connect_dataset_interface(dataset_interface, data_loader_num_workers)
        return self.dataset_interface

    def connect_dataset_interface(self, dataset_interface, data_loader_num_workers):
        self.dataset_interface = dataset_interface
        self.train_loader, self.valid_loader, self.test_loader, self.classes = \
            self.dataset_interface.get_data_loaders(batch_size_factor=self.num_devices,
                                                    num_workers=data_loader_num_workers,
                                                    distributed_sampler=self.multi_gpu == MultiGPUMode.DISTRIBUTED_DATA_PARALLEL)

        self.dataset_params = self.dataset_interface.get_dataset_params()

    def _add_metrics_update_callback(self, phase):
        """
        Adds AuxModelMetricsUpdateCallback to be fired at phase

        :param phase: Phase for the metrics callback to be fired at
        """
        self.phase_callbacks.append(AuxMetricsUpdateCallback(phase))

    def _load_model(self, params):
        model_params = params['model']
        input_channels = len(params['dataset']['channels'])
        output_channels = params['dataset']['num_classes']
        arch_params = {
            'input_channels': input_channels,
            'output_channels': output_channels,
            'in_channels': input_channels,
            'out_channels': output_channels,
            'num_classes': output_channels,
            **model_params['params']
        }
        try:
            model = instantiate_class(model_params['name'], arch_params)
        except (AttributeError, ValueError):
            if model_params['name'] in MODELS_DICT.keys():
                model = MODELS_DICT[model_params['name']](arch_params)
            else:
                model = models.get(model_name=model_params['name'], arch_params=arch_params)

        return model, arch_params

    def init_model(self, params: Mapping, resume: bool, checkpoint_path: str = None):
        # load model
        model, arch_params = self._load_model(params)

        if 'num_classes' not in arch_params.keys():
            if self.classes is None and self.dataset_interface is None:
                raise Exception('Error', 'Number of classes not defined in arch params and dataset is not defined')
            else:
                arch_params['num_classes'] = len(self.classes)

        self.arch_params = core_utils.HpmStruct(**arch_params)
        self.net = model

        self._net_to_device()
        # SET THE FLAG FOR DIFFERENT PARAMETER GROUP OPTIMIZER UPDATE
        self.update_param_groups = hasattr(self.net.module, 'update_param_groups')

        if resume and checkpoint_path is not None:
            self.checkpoint = load_checkpoint_to_model(ckpt_local_path=checkpoint_path,
                                                       load_backbone=False,
                                                       net=self.net,
                                                       strict=StrictLoad.ON.value,
                                                       load_weights_only=self.load_weights_only)
            self.load_checkpoint = True

            if 'ema_net' in self.checkpoint.keys():
                logger.warning(
                    "[WARNING] Main network has been loaded from checkpoint but EMA network exists as well. It "
                    " will only be loaded during validation when training with ema=True. ")

            # UPDATE TRAINING PARAMS IF THEY EXIST & WE ARE NOT LOADING AN EXTERNAL MODEL's WEIGHTS
            self.best_metric = self.checkpoint['acc'] if 'acc' in self.checkpoint.keys() else -1
            self.start_epoch = self.checkpoint['epoch'] if 'epoch' in self.checkpoint.keys() else 0

    def train(self, training_params: dict = dict()):
        super().train(model=self.net,
                      training_params=training_params,
                      train_loader=self.train_loader,
                      valid_loader=self.valid_loader)
        if self.train_loader.num_workers > 0 and self.train_loader._iterator is not None:
            self.train_loader._iterator._shutdown_workers()
        if self.valid_loader.num_workers > 0 and self.valid_loader._iterator is not None:
            self.valid_loader._iterator._shutdown_workers()
        self._restore_best_params()

    def _train_epoch(self, epoch: int, silent_mode: bool = False) -> tuple:
        """
        train_epoch - A single epoch training procedure
            :param optimizer:   The optimizer for the network
            :param epoch:       The current epoch
            :param silent_mode: No verbosity
        """
        # SET THE MODEL IN training STATE
        self.net.train()
        # THE DISABLE FLAG CONTROLS WHETHER THE PROGRESS BAR IS SILENT OR PRINTS THE LOGS
        progress_bar_train_loader = tqdm(self.train_loader, bar_format="{l_bar}{bar:10}{r_bar}", dynamic_ncols=True, disable=silent_mode)
        progress_bar_train_loader.set_description(f"Train epoch {epoch}")

        # RESET/INIT THE METRIC LOGGERS
        self._reset_metrics()

        self.train_metrics.to(device_config.device)
        loss_avg_meter = core_utils.utils.AverageMeter()

        context = PhaseContext(
            epoch=epoch,
            optimizer=self.optimizer,
            metrics_compute_fn=self.train_metrics,
            loss_avg_meter=loss_avg_meter,
            criterion=self.criterion,
            device=device_config.device,
            lr_warmup_epochs=self.training_params.lr_warmup_epochs,
            sg_logger=self.sg_logger,
            train_loader=self.train_loader,
            context_methods=self._get_context_methods(Phase.TRAIN_BATCH_END),
            ddp_silent_mode=self.ddp_silent_mode,
        )

        for batch_idx, batch_items in enumerate(progress_bar_train_loader):
            batch_items = core_utils.tensor_container_to_device(batch_items, device_config.device, non_blocking=True)
            inputs, targets, additional_batch_items = sg_trainer_utils.unpack_batch_items(batch_items)

            if self.pre_prediction_callback is not None:
                inputs, targets = self.pre_prediction_callback(inputs, targets, batch_idx)

            context.update_context(batch_idx=batch_idx, inputs=inputs, target=targets, **additional_batch_items)
            self.phase_callback_handler.on_train_batch_start(context)

            # AUTOCAST IS ENABLED ONLY IF self.training_params.mixed_precision - IF enabled=False AUTOCAST HAS NO EFFECT
            with autocast(enabled=self.training_params.mixed_precision):
                # FORWARD PASS TO GET NETWORK'S PREDICTIONS
                outputs = self._forward_step(inputs, targets, additional_batch_items)

                # COMPUTE THE LOSS FOR BACK PROP + EXTRA METRICS COMPUTED DURING THE LOSS FORWARD PASS
                loss, loss_log_items = self._get_losses(outputs, targets)

            context.update_context(preds=outputs, loss_log_items=loss_log_items)
            self.phase_callback_handler.on_train_batch_loss_end(context)

            # LOG LR THAT WILL BE USED IN CURRENT EPOCH AND AFTER FIRST WARMUP/LR_SCHEDULER UPDATE BEFORE WEIGHT UPDATE
            if not self.ddp_silent_mode and batch_idx == 0:
                self._write_lrs(epoch)

            self._backward_step(loss, epoch, batch_idx, context)

            # COMPUTE THE RUNNING USER METRICS AND LOSS RUNNING ITEMS. RESULT TUPLE IS THEIR CONCATENATION.
            logging_values = loss_avg_meter.average + get_metrics_results_tuple(self.train_metrics)
            gpu_memory_utilization = get_gpu_mem_utilization() / 1e9 if torch.cuda.is_available() else 0

            # RENDER METRICS PROGRESS
            pbar_message_dict = get_train_loop_description_dict(
                logging_values, self.train_metrics, self.loss_logging_items_names, gpu_mem=gpu_memory_utilization
            )

            progress_bar_train_loader.set_postfix(**pbar_message_dict)
            self.phase_callback_handler.on_train_batch_end(context)

            # TODO: ITERATE BY MAX ITERS
            # FOR INFINITE SAMPLERS WE MUST BREAK WHEN REACHING LEN ITERATIONS.
            if (self._infinite_train_loader and batch_idx == len(self.train_loader) - 1) or (
                self.max_train_batches is not None and self.max_train_batches - 1 <= batch_idx
            ):
                break

        self.train_monitored_values = sg_trainer_utils.update_monitored_values_dict(
            monitored_values_dict=self.train_monitored_values, new_values_dict=pbar_message_dict
        )

        return logging_values
    
    def _forward_step(self, inputs, targets, additional_batch_items):
        if core_utils.get_param(self.training_params, "pass_targets_to_net", False):
            if core_utils.get_param(self.training_params, "pass_additional_batch_items_to_net", False):
                return self.net(inputs, targets, additional_batch_items)
            return self.net(inputs, targets)
        if core_utils.get_param(self.training_params, "pass_additional_batch_items_to_net", False):
            return self.net(inputs, additional_batch_items)
        return self.net(inputs)
    
    def _get_losses(self, outputs: torch.Tensor, targets: torch.Tensor) -> Tuple[torch.Tensor, tuple]:
        # GET THE OUTPUT OF THE LOSS FUNCTION
        if self.validation and core_utils.get_param(self.training_params, "inform_loss_in_validaiton", False):
            loss = self.criterion(outputs, targets, validation=self.validation)
        else:
            loss = self.criterion(outputs, targets)
        if isinstance(loss, tuple):
            loss, loss_logging_items = loss
            # IF ITS NOT A TUPLE THE LOGGING ITEMS CONTAIN ONLY THE LOSS FOR BACKPROP (USER DEFINED LOSS RETURNS SCALAR)
        else:
            loss_logging_items = loss.unsqueeze(0).detach()

        # ON FIRST BACKWARD, DERRIVE THE LOGGING TITLES.
        if self.loss_logging_items_names is None or self._first_backward:
            self._init_loss_logging_names(loss_logging_items)
            if self.metric_to_watch:
                self._init_monitored_items()
            self._first_backward = False

        if len(loss_logging_items) != len(self.loss_logging_items_names):
            raise ValueError(
                "Loss output length must match loss_logging_items_names. Got "
                + str(len(loss_logging_items))
                + ", and "
                + str(len(self.loss_logging_items_names))
            )
        # RETURN AND THE LOSS LOGGING ITEMS COMPUTED DURING LOSS FORWARD PASS
        return loss, loss_logging_items

    def _validate_epoch(self, epoch: int, silent_mode: bool = False) -> tuple:
        self.validation = True
        res = super()._validate_epoch(epoch, silent_mode)
        self.validation = False
        return res
    
    def evaluate(
        self,
        data_loader: torch.utils.data.DataLoader,
        metrics: MetricCollection,
        evaluation_type: EvaluationType,
        epoch: int = None,
        silent_mode: bool = False,
        metrics_progress_verbose: bool = False,
    ):
        """
        Evaluates the model on given dataloader and metrics.

        :param data_loader: dataloader to perform evaluataion on
        :param metrics: (MetricCollection) metrics for evaluation
        :param evaluation_type: (EvaluationType) controls which phase callbacks will be used (for example, on batch end,
            when evaluation_type=EvaluationType.VALIDATION the Phase.VALIDATION_BATCH_END callbacks will be triggered)
        :param epoch: (int) epoch idx
        :param silent_mode: (bool) controls verbosity
        :param metrics_progress_verbose: (bool) controls the verbosity of metrics progress (default=False).
            Slows down the program significantly.

        :return: results tuple (tuple) containing the loss items and metric values.
        """

        # THE DISABLE FLAG CONTROLS WHETHER THE PROGRESS BAR IS SILENT OR PRINTS THE LOGS
        progress_bar_data_loader = tqdm(data_loader, bar_format="{l_bar}{bar:10}{r_bar}", dynamic_ncols=True, disable=silent_mode)
        loss_avg_meter = core_utils.utils.AverageMeter()
        logging_values = None
        loss_tuple = None
        lr_warmup_epochs = self.training_params.lr_warmup_epochs if self.training_params else None
        context = PhaseContext(
            epoch=epoch,
            metrics_compute_fn=metrics,
            loss_avg_meter=loss_avg_meter,
            criterion=self.criterion,
            device=device_config.device,
            lr_warmup_epochs=lr_warmup_epochs,
            sg_logger=self.sg_logger,
            context_methods=self._get_context_methods(Phase.VALIDATION_BATCH_END),
        )

        if not silent_mode:
            # PRINT TITLES
            pbar_start_msg = f"Validation epoch {epoch}" if evaluation_type == EvaluationType.VALIDATION else "Test"
            progress_bar_data_loader.set_description(pbar_start_msg)

        with torch.no_grad():
            for batch_idx, batch_items in enumerate(progress_bar_data_loader):
                batch_items = core_utils.tensor_container_to_device(batch_items, device_config.device, non_blocking=True)
                inputs, targets, additional_batch_items = sg_trainer_utils.unpack_batch_items(batch_items)

                # TRIGGER PHASE CALLBACKS CORRESPONDING TO THE EVALUATION TYPE
                context.update_context(batch_idx=batch_idx, inputs=inputs, target=targets, **additional_batch_items)
                if evaluation_type == EvaluationType.VALIDATION:
                    self.phase_callback_handler.on_validation_batch_start(context)
                else:
                    self.phase_callback_handler.on_test_batch_start(context)

                output = self._forward_step(inputs, targets, batch_items)
                context.update_context(preds=output)

                if self.criterion is not None:
                    # STORE THE loss_items ONLY, THE 1ST RETURNED VALUE IS THE loss FOR BACKPROP DURING TRAINING
                    loss_tuple = self._get_losses(output, targets)[1].cpu()
                    context.update_context(loss_log_items=loss_tuple)

                # TRIGGER PHASE CALLBACKS CORRESPONDING TO THE EVALUATION TYPE
                if evaluation_type == EvaluationType.VALIDATION:
                    self.phase_callback_handler.on_validation_batch_end(context)
                else:
                    self.phase_callback_handler.on_test_batch_end(context)

                # COMPUTE METRICS IF PROGRESS VERBOSITY IS SET
                if metrics_progress_verbose and not silent_mode:
                    # COMPUTE THE RUNNING USER METRICS AND LOSS RUNNING ITEMS. RESULT TUPLE IS THEIR CONCATENATION.
                    logging_values = get_logging_values(loss_avg_meter, metrics, self.criterion)
                    pbar_message_dict = get_train_loop_description_dict(logging_values, metrics, self.loss_logging_items_names)

                    progress_bar_data_loader.set_postfix(**pbar_message_dict)

                if evaluation_type == EvaluationType.VALIDATION and self.max_valid_batches is not None and self.max_valid_batches - 1 <= batch_idx:
                    break

        # NEED TO COMPUTE METRICS FOR THE FIRST TIME IF PROGRESS VERBOSITY IS NOT SET
        if not metrics_progress_verbose:
            # COMPUTE THE RUNNING USER METRICS AND LOSS RUNNING ITEMS. RESULT TUPLE IS THEIR CONCATENATION.
            logging_values = get_logging_values(loss_avg_meter, metrics, self.criterion)
            pbar_message_dict = get_train_loop_description_dict(logging_values, metrics, self.loss_logging_items_names)

            progress_bar_data_loader.set_postfix(**pbar_message_dict)

        # TODO: SUPPORT PRINTING AP PER CLASS- SINCE THE METRICS ARE NOT HARD CODED ANYMORE (as done in
        #  calc_batch_prediction_accuracy_per_class in metric_utils.py), THIS IS ONLY RELEVANT WHEN CHOOSING
        #  DETECTIONMETRICS, WHICH ALREADY RETURN THE METRICS VALUEST HEMSELVES AND NOT THE ITEMS REQUIRED FOR SUCH
        #  COMPUTATION. ALSO REMOVE THE BELOW LINES BY IMPLEMENTING CRITERION AS A TORCHMETRIC.

        if device_config.multi_gpu == MultiGPUMode.DISTRIBUTED_DATA_PARALLEL:
            logging_values = reduce_results_tuple_for_ddp(logging_values, next(self.net.parameters()).device)

        pbar_message_dict = get_train_loop_description_dict(logging_values, metrics, self.loss_logging_items_names)

        self.valid_monitored_values = sg_trainer_utils.update_monitored_values_dict(
            monitored_values_dict=self.valid_monitored_values, new_values_dict=pbar_message_dict
        )

        if not silent_mode and evaluation_type == EvaluationType.VALIDATION:
            progress_bar_data_loader.write("===========================================================")
            sg_trainer_utils.display_epoch_summary(
                epoch=context.epoch, n_digits=4, train_monitored_values=self.train_monitored_values, valid_monitored_values=self.valid_monitored_values
            )
            progress_bar_data_loader.write("===========================================================")

        return logging_values

    def _net_to_device(self):
        """
        Manipulates self.net according to device.multi_gpu
        """
        self.net.to(device_config.device)

        # FOR MULTI-GPU TRAINING (not distributed)
        sync_bn = core_utils.get_param(self.training_params, "sync_bn", default_val=False)
        if device_config.multi_gpu == MultiGPUMode.DATA_PARALLEL:
            self.net = torch.nn.DataParallel(self.net, device_ids=get_device_ids())
        elif device_config.multi_gpu == MultiGPUMode.DISTRIBUTED_DATA_PARALLEL:
            if sync_bn:
                if not self.ddp_silent_mode:
                    logger.info("DDP - Using Sync Batch Norm... Training time will be affected accordingly")
                self.net = torch.nn.SyncBatchNorm.convert_sync_batchnorm(self.net).to(device_config.device)

            local_rank = int(device_config.device.split(":")[1])
            self.net = torch.nn.parallel.DistributedDataParallel(self.net, device_ids=[local_rank], output_device=local_rank, find_unused_parameters=True)

        elif not isinstance(self.net, WrappedModel):
            self.net = WrappedModel(self.net)
        else:
            pass
        
    def _restore_best_params(self):
        """
        Restore best parameters after the training
        """
        self.checkpoint = load_checkpoint_to_model(ckpt_local_path=self.checkpoints_dir_path + '/ckpt_best.pth',
                                                   load_backbone=False,
                                                   net=self.net,
                                                   strict=StrictLoad.ON.value,
                                                   load_weights_only=True)

    def test(self,  # noqa: C901
             test_loader: torch.utils.data.DataLoader = None,
             loss: torch.nn.modules.loss._Loss = None,
             silent_mode: bool = False,
             test_metrics: Mapping = None,
             loss_logging_items_names=None, metrics_progress_verbose=False, test_phase_callbacks={},
             use_ema_net=True) -> dict:
        """
        Evaluates the model on given dataloader and metrics.

        :param test_loader: dataloader to perform test on.
        :param test_metrics: dict name: Metric for evaluation.
        :param silent_mode: (bool) controls verbosity
        :param metrics_progress_verbose: (bool) controls the verbosity of metrics progress (default=False). Slows down the program.
        :param use_ema_net (bool) whether to perform test on self.ema_model.ema (when self.ema_model.ema exists,
            otherwise self.net will be tested) (default=True)
        :return: results tuple (tuple) containing the loss items and metric values.

        All of the above args will override SgModel's corresponding attribute when not equal to None. Then evaluation
         is ran on self.test_loader with self.test_metrics.
        """
        test_loader = test_loader or self.test_loader
        loss = loss or self.training_params.loss
        self.validation = True
        if loss is not None:
            loss.to(self.device)
        test_phase_callbacks = [callback_factory(name, params, seg_trainer=self, dataset=self.dataset_interface, loader=test_loader)
                                for name, params in test_phase_callbacks.items()]
        metrics_values = super().test(model=self.net,
                                      test_loader=test_loader,
                                      loss=loss,
                                      silent_mode=silent_mode,
                                      test_metrics_list=list(test_metrics.values()),
                                      loss_logging_items_names=loss_logging_items_names,
                                      metrics_progress_verbose=metrics_progress_verbose,
                                      test_phase_callbacks=test_phase_callbacks,
                                      use_ema_net=use_ema_net)
        self.validation = False


        if self.test_loader.num_workers > 0:
            self.test_loader._iterator._shutdown_workers()

        metric_names = test_metrics.keys()
        loss_names = self.loss_logging_items_names
        if len(loss_names) + len(metric_names) != len(metrics_values):
            raise ValueError(f'Number of loss names ({len(loss_names)}) + number of metric names ({len(metric_names)}) '
                             f'!= number of metrics values ({len(metrics_values)})')
        losses = dict(zip(loss_names, metrics_values[:len(loss_names)]))
        metrics = dict(zip(metric_names, metrics_values[len(loss_names):]))
        metrics = {**losses, **metrics}

        if 'conf_mat' in metrics:
            metrics.pop('conf_mat')
            cf = test_metrics['conf_mat'].get_cf()
            logger.info(f'Confusion matrix:\n{cf}')
            self.sg_logger.add_table('confusion_matrix', cf.cpu(),
                                     columns=list(self.dataset_interface.testset.CLASS_LABELS.values()),
                                     rows=list(self.dataset_interface.testset.CLASS_LABELS.values())
                                     )
        self.sg_logger.add_summary(metrics)
        logger.info(f'Test metrics: {pformat(metrics)}')
        if 'auc' in metrics:
            logger.info('Computing ROC curve...')
            roc = test_metrics['auc'].get_roc()
            fpr_tpr = [(roc[0][i], roc[1][i]) for i in range(len(roc))]
            skip = [x[0].shape.numel() // 1000 for x in fpr_tpr]
            fpr_tpr = [(fpr[::sk], tpr[::sk]) for (fpr, tpr), sk in zip(fpr_tpr, skip)]
            fprs, tprs = zip(*fpr_tpr)
            fprs = torch.cat(fprs)
            tprs = torch.cat(tprs)
            classes = list(self.dataset_interface.testset.CLASS_LABELS.values())
            cls = [[classes[i]] * len(fpr)
                   for i, (fpr, tpr) in enumerate(fpr_tpr)]
            cls = [item for sublist in cls for item in sublist]
            df = pd.DataFrame({'class': cls, 'fpr': fprs, 'tpr': tprs})
            name = "ROCAUC"
            self.sg_logger.add_plot(name, df, "fpr", "tpr", classes_marker="class")
            logger.info('ROC curve computed.')
        return metrics

    def _init_monitored_items(self):
        if self.metric_to_watch == 'loss':
            if len(self.loss_logging_items_names) == 1:
                self.metric_to_watch = self.loss_logging_items_names[0]
            else:
                raise ValueError(f"Specify which loss {self.loss_logging_items_names} to watch")
        return super()._init_monitored_items()

    def _initialize_sg_logger_objects(self, additional_configs_to_log: Dict = None):
        if not self.train_initialized:
            self.train_initialized = True
            # OVERRIDE SOME PARAMETERS TO MAKE SURE THEY MATCH THE TRAINING PARAMETERS
            general_sg_logger_params = {  # 'experiment_name': self.experiment_name,
                'experiment_name': '',
                'group': self.experiment_name,
                'storage_location': self.model_checkpoints_location,
                'resumed': self.load_checkpoint,
                'training_params': self.training_params,
                'checkpoints_dir_path': self.checkpoints_dir_path,
                'run_id': self.run_id
            }
            sg_logger = core_utils.get_param(self.training_params, 'sg_logger')

            if sg_logger is None:
                raise RuntimeError('logger must be defined in experiment params (see default_training_params)')

            if isinstance(sg_logger, AbstractSGLogger):
                self.sg_logger = sg_logger
            elif isinstance(sg_logger, str):
                sg_logger_params = {**general_sg_logger_params,
                                    **core_utils.get_param(self.training_params, 'sg_logger_params', {})}
                if sg_logger in SG_LOGGERS:
                    self.sg_logger = SG_LOGGERS[sg_logger](**sg_logger_params)
                elif sg_logger in LOGGERS:
                    self.sg_logger = LOGGERS[sg_logger](**sg_logger_params)
                else:
                    raise RuntimeError('sg_logger not defined in SG_LOGGERS of SuperGradients neither in LOGGERS of '
                                       'EzDL')
            else:
                raise RuntimeError('sg_logger can be either an sg_logger name (str) or a subcalss of AbstractSGLogger')

            if not (isinstance(self.sg_logger, BaseSGLogger) or isinstance(self.sg_logger, BaseLogger)):
                logger.warning(
                    "WARNING! Using a user-defined sg_logger: files will not be automatically written to disk!\n"
                    "Please make sure the provided sg_logger writes to disk or compose your sg_logger to BaseSGLogger")

            # IN CASE SG_LOGGER UPDATED THE DIR PATH
            self.checkpoints_dir_path = self.sg_logger.local_dir()
            self.training_params.override(sg_logger=sg_logger)
            additional_log_items = {'num_devices': self.num_devices,
                                    'multi_gpu': str(self.multi_gpu),
                                    'device_type': torch.cuda.get_device_name(
                                        0) if torch.cuda.is_available() else 'cpu'}

            # ADD INSTALLED PACKAGE LIST + THEIR VERSIONS
            if self.training_params.log_installed_packages:
                pkg_list = list(map(lambda pkg: str(pkg), _get_installed_distributions()))
                additional_log_items['installed_packages'] = pkg_list

            self.sg_logger.add_config("additional_log_items", additional_log_items)
            self.sg_logger.flush()

    def init_loggers(self,
                     in_params: Mapping = None,
                     train_params: Mapping = None,
                     init_sg_loggers: bool = True,
                     run_id=None) -> None:

        self.run_id = run_id
        if self.training_params is None:
            self.training_params = TrainingParams()
        self.training_params.override(**train_params)
        if init_sg_loggers:
            self._initialize_sg_logger_objects()
        if self.phase_callbacks is None:
            self.phase_callbacks = []
        self.phase_callback_handler = CallbackHandler(self.phase_callbacks)
        self.sg_logger.add_config(config=in_params)

    def _save_checkpoint(self, optimizer=None, epoch: int = None, validation_results_tuple: tuple = None, context: PhaseContext = None):
        """
        Save the current state dict as latest (always), best (if metric was improved), epoch# (if determined in training
        params)
        """
        # WHEN THE validation_results_tuple IS NONE WE SIMPLY SAVE THE state_dict AS LATEST AND Return
        if validation_results_tuple is None:
            self.sg_logger.add_checkpoint(tag="ckpt_latest_weights_only.pth", state_dict={"net": self.net.state_dict()}, global_step=epoch)
            return

        # COMPUTE THE CURRENT metric
        # IF idx IS A LIST - SUM ALL THE VALUES STORED IN THE LIST'S INDICES
        metric = (
            validation_results_tuple[self.metric_idx_in_results_tuple]
            if isinstance(self.metric_idx_in_results_tuple, int)
            else sum([validation_results_tuple[idx] for idx in self.metric_idx_in_results_tuple])
        )

        # BUILD THE state_dict
        state = {"net": self.net.state_dict(), "acc": metric, "epoch": epoch}
        if optimizer is not None:
            state["optimizer_state_dict"] = optimizer.state_dict()

        if self.scaler is not None:
            state["scaler_state_dict"] = self.scaler.state_dict()

        if self.ema:
            state["ema_net"] = self.ema_model.ema.state_dict()
        # SAVES CURRENT MODEL AS ckpt_latest
        self.sg_logger.add_checkpoint(tag="ckpt_latest.pth", state_dict=state, global_step=epoch)

        # SAVE MODEL AT SPECIFIC EPOCHS DETERMINED BY save_ckpt_epoch_list
        if epoch in self.training_params.save_ckpt_epoch_list:
            self.sg_logger.add_checkpoint(tag=f"ckpt_epoch_{epoch}.pth", state_dict=state, global_step=epoch)

        # OVERRIDE THE BEST CHECKPOINT AND best_metric IF metric GOT BETTER THAN THE PREVIOUS BEST
        if (metric > self.best_metric and self.greater_metric_to_watch_is_better) or (metric < self.best_metric and not self.greater_metric_to_watch_is_better):
            # STORE THE CURRENT metric AS BEST
            self.best_metric = metric
            self._save_best_checkpoint(epoch, state)

            # RUN PHASE CALLBACKS
            self.phase_callback_handler.on_validation_end_best_epoch(context)

            if isinstance(metric, torch.Tensor):
                metric = metric.item()
            logger.info("Best checkpoint overriden: validation " + self.metric_to_watch + ": " + str(metric))

        if self.training_params.average_best_models:
            net_for_averaging = self.ema_model.ema if self.ema else self.net
            state["net"] = self.model_weight_averaging.get_average_model(net_for_averaging, validation_results_tuple=validation_results_tuple)
            self.sg_logger.add_checkpoint(tag=self.average_model_checkpoint_filename, state_dict=state, global_step=epoch)

    def _save_best_checkpoint(self, epoch, state):
        if self.ema:
            best_net = self.ema_model.ema
            state.pop("ema_net")
        else:
            best_net = self.net

        state["net"] = best_net.state_dict()
        self.sg_logger.add_checkpoint(tag=self.ckpt_best_name, state_dict=state, global_step=epoch)

    def run(self, data_loader: torch.utils.data.DataLoader, callbacks=None, silent_mode: bool = False):
        """
        Runs the model on given dataloader.

        :param data_loader: dataloader to perform run on

        """

        # THE DISABLE FLAG CONTROLS WHETHER THE PROGRESS BAR IS SILENT OR PRINTS THE LOGS
        if callbacks is None:
            callbacks = []
        progress_bar_data_loader = tqdm(data_loader, bar_format="{l_bar}{bar:10}{r_bar}", dynamic_ncols=True,
                                        disable=silent_mode)
        context = PhaseContext(criterion=self.criterion,
                               device=self.device,
                               sg_logger=self.sg_logger)

        self.phase_callbacks.extend(callbacks)

        if not silent_mode:
            # PRINT TITLES
            pbar_start_msg = f"Running model on {len(data_loader)} batches"
            progress_bar_data_loader.set_description(pbar_start_msg)

        with torch.no_grad():
            for batch_idx, batch_items in enumerate(progress_bar_data_loader):
                batch_items = core_utils.tensor_container_to_device(batch_items, self.device, non_blocking=True)

                additional_batch_items = {}
                targets = None
                if hasattr(batch_items, '__len__'):
                    if len(batch_items) == 2:
                        inputs, targets = batch_items
                    elif len(batch_items) == 3:
                        inputs, targets, additional_batch_items = batch_items
                    else:
                        raise ValueError(f"Expected 1, 2 or 3 items in batch_items, got {len(batch_items)}")
                else:
                    inputs = batch_items

                output = self.net(inputs)

                context.update_context(batch_idx=batch_idx,
                                       inputs=inputs,
                                       preds=output,
                                       target=targets,
                                       **additional_batch_items)

                # TRIGGER PHASE CALLBACKS
                self.phase_callback_handler(Phase.POST_TRAINING, context)





class PolyLR:
    def __init__(self, power):
        self.poly_exp = power

    def perform_scheduling(self, initial_lr, epoch, max_epoch, **kwargs):
        """
        Perform polyLR learning rate scheduling
        """
        return initial_lr * pow((1.0 - epoch / max_epoch), self.poly_exp)


class ExpLR:
    def __init__(self, gamma, last_epoch_decay=None, min_lr=None):
        self.gamma = gamma
        self.last_epoch_decay = last_epoch_decay
        self.min_lr = min_lr

    def perform_scheduling(self, initial_lr, epoch, **kwargs):
        """
        Perform polyLR learning rate scheduling
        """
        if (
            self.min_lr is not None
            and initial_lr * pow(self.gamma, epoch) < self.min_lr
        ):
            return self.min_lr
        if self.last_epoch_decay is not None and epoch >= self.last_epoch_decay:
            return initial_lr * pow(self.gamma, self.last_epoch_decay)
        return initial_lr * pow(self.gamma, epoch)



from collections import namedtuple
from inspect import signature

import torch
from super_gradients.training.models import SgModule
from super_gradients.training.utils import HpmStruct
from torchdistill.core.forward_hook import ForwardHookManager

from torch import nn

from ezdl.models.layers.common import ConvModule
from ezdl.utils.sg import unwrap_model_from_parallel
from ezdl.models.kd.logits import LogitsDistillationModule


FDOutput = namedtuple('FDOutput', ['student_features',
                                    'student_output',
                                    'teacher_features',
                                    'teacher_output'])


class FeatureDistillationModule(LogitsDistillationModule):
    """
    Feature Distillation Module
    """
    def __init__(self, arch_params: HpmStruct, student: SgModule, teacher: torch.nn.Module, run_teacher_on_eval=False):
        """
        :param arch_params: architecture parameters
        :param student: student model
        :param teacher: teacher model
        :param run_teacher_on_eval: whether to run the teacher on eval
        """
        super().__init__(arch_params=arch_params,
                         student=student, teacher=teacher,
                         run_teacher_on_eval=run_teacher_on_eval)

        if 'hooks' in arch_params:
            self.teacher_hook_manager = ForwardHookManager(next(self.teacher.parameters()).device)
            self.student_hook_manager = ForwardHookManager(next(self.student.parameters()).device)
            self.add_hooks(arch_params['hooks'])
            self.hooks = True
        else:
            self.teacher_hook_manager = None
            self.student_hook_manager = None
            student, swrapped = unwrap_model_from_parallel(student, return_was_wrapped=True)
            teacher, twrapped = unwrap_model_from_parallel(teacher, return_was_wrapped=True)
            assert 'return_encoding' in signature(student.forward).parameters, \
                "Student model must return encoding if no hooks are used"
            assert 'return_encoding' in signature(teacher.forward).parameters, \
                "Teacher model must return encoding if no hooks are used"
            self.hooks = False
            if swrapped:
                self.student.forward = self.student.module.forward
            if twrapped:
                self.teacher.forward = self.teacher.module.forward

    def add_hooks(self, hooks):
        """
        Add hooks to the teacher and student
        :param hooks: list of hooks
        """
        if isinstance(hooks, dict):
            student_hooks = hooks['student']
            teacher_hooks = hooks['teacher']
        else:
            student_hooks = hooks
            teacher_hooks = hooks
        self._add_hooks_to_model(self.student, self.student_hook_manager, student_hooks)
        self._add_hooks_to_model(self.teacher, self.teacher_hook_manager, teacher_hooks)


    def _add_hooks_to_model(self, model, hook_manager, hooks):
        """
        Add hooks to a model
        :param model: model to add hooks to
        """
        model = unwrap_model_from_parallel(model)
        for hook in hooks:
            if isinstance(hook, str):
                hook_manager.add_hook(model, hook, requires_output=True)
            elif isinstance(hook, dict):
                hook_manager.add_hook(model, hook['name'], **hook['params'])
            else:
                raise ValueError("Invalid hook")

    def forward(self, x):
        """
        Forward pass
        :param x: input
        :return: output
        """
        return self.hook_forward(x) if self.hooks else self.stepped_forward(x)

    def hook_forward(self, x):
        student_output = self.student.forward(x)
        if self.teacher_input_adapter is not None:
            teacher_output = self.teacher.forward(self.teacher_input_adapter(x))
        else:
            teacher_output = self.teacher.forward(x)
        student_features = self.student_hook_manager.pop_io_dict()
        student_features = [student_features[hook]['output'] for hook in student_features]
        teacher_features = self.teacher_hook_manager.pop_io_dict()
        teacher_features = [teacher_features[hook]['output'] for hook in teacher_features]
    
        return FDOutput(
            student_features=student_features,
            student_output=student_output,
            teacher_features=teacher_features,
            teacher_output=teacher_output
        )

    def stepped_forward(self, x):
        """
        Forward pass for a single step
        :param x: input
        :return: output
        """
        student_output, student_features = self.student.forward(x, return_encoding=True)
        if self.teacher_input_adapter is not None:
            teacher_output, teacher_features = self.teacher.forward(self.teacher_input_adapter(x), return_encoding=True)
        else:
            teacher_output, teacher_features = self.teacher.forward(x, return_encoding=True)
        return FDOutput(
            student_features=student_features,
            student_output=student_output,
            teacher_features=teacher_features,
            teacher_output=teacher_output
        )


class FeatureDistillationConvAdapter(FeatureDistillationModule):
    def __init__(self, arch_params: HpmStruct, student: SgModule, teacher: torch.nn.Module, run_teacher_on_eval=False,
                 epsilon=1e-6):
        super().__init__(arch_params=arch_params,
                         student=student, teacher=teacher,
                         run_teacher_on_eval=run_teacher_on_eval)
        self.adapters = nn.ModuleList([ConvModule(inp, out, 3, p="same") for inp, out in
                            zip(self.student.module.encoder_maps_sizes, self.teacher.module.encoder_maps_sizes)])
    def forward(self, x):
        fd_output = super().forward(x)
        student_features = [adapter(feat_map) for feat_map, adapter in zip(fd_output.student_features, self.adapters)]
        return FDOutput(
            student_features=student_features,
            student_output=fd_output.student_output,
            teacher_features=fd_output.teacher_features,
            teacher_output=fd_output.teacher_output
        )

class VariationalInformationDistillation(FeatureDistillationModule):
    """
    Feature Distillation Module that uses Variational Information
    """
    def __init__(self, arch_params: HpmStruct, student: SgModule, teacher: torch.nn.Module, run_teacher_on_eval=False,
                 epsilon=1e-6):
        super().__init__(arch_params=arch_params,
                         student=student, teacher=teacher,
                         run_teacher_on_eval=run_teacher_on_eval)
        self.mu_networks = nn.ModuleList([ConvModule(inp, out, 3, p="same") for inp, out in
                            zip(self.student.module.encoder_maps_sizes, self.teacher.module.encoder_maps_sizes)])
        self.alphas = nn.ParameterList([torch.nn.Parameter(torch.rand(1)) for _ in self.mu_networks])
        self.sigmas = nn.ModuleList([SigmaVariance(epsilon) for _ in self.mu_networks])

    def forward(self, x):
        fd_output = super().forward(x)
        mu = [mu_network(feat_map) for feat_map, mu_network in zip(fd_output.student_features, self.mu_networks)]
        sigmas = [sigma() for sigma in self.sigmas]
        return FDOutput(
            student_features=list(zip(mu, sigmas)),
            student_output=fd_output.student_output,
            teacher_features=fd_output.teacher_features,
            teacher_output=fd_output.teacher_output
        )


class SigmaVariance(nn.Module):
    def __init__(self, epsilon=1e-3):
        super().__init__()
        self.epsilon = epsilon
        self.sigma = nn.Parameter(torch.rand(1), requires_grad=True)

    def forward(self):
        return torch.nn.functional.softplus(self.sigma) + self.epsilon



import gc
import sys
import os
import traceback

from copy import deepcopy
from requests.exceptions import ConnectionError

from super_gradients.training.utils.callbacks import Phase
from codecarbon import EmissionsTracker, OfflineEmissionsTracker
from ezdl.logger.basesg_logger import AbstractRunWrapper

from ezdl.logger.text_logger import get_logger
from ezdl.callbacks import MetricsLogCallback, callback_factory
from ezdl.experiment.kd_seg_trainer import KDSegTrainer
from ezdl.experiment.ez_trainer import EzTrainer
from ezdl.experiment.kd_ez_trainer import KDEzTrainer
from ezdl.experiment.parameters import parse_params
from ezdl.experiment.seg_trainer import SegmentationTrainer
from ezdl.utils.utilities import dict_to_yaml_string, values_to_number, nested_dict_update

logger = get_logger(__name__)


class Run:
    def __init__(self):
        self.kd = None
        self.params = None
        self.dataset = None
        self.train_callbacks = None
        self.val_callbacks = None
        self.test_callbacks = None
        self.dataset_params = None
        self.seg_trainer = None
        self.train_params = None
        self.test_params = None
        self.run_params = None
        self.phases = None
        self.carbon_tracker = None
        if '.' not in sys.path:
            sys.path.extend('.')

    def parse_params(self, params):
        self.params = deepcopy(params)
        self.phases = params['phases']

        self.train_params, self.test_params, self.dataset_params, callbacks, self.kd = parse_params(self.params)
        self.train_callbacks, self.val_callbacks, self.test_callbacks = callbacks
        self.run_params = params.get('run_params') or {}
        
    def _init_carbon_tracker(self):
        try:
            self.carbon_tracker = EmissionsTracker(output_dir=self.seg_trainer.sg_logger._local_dir, log_level="warning")
        except ConnectionError:
            logger.warning("CodeCarbon is not connected to a server, using offline tracker")
            self.carbon_tracker = OfflineEmissionsTracker(output_dir=self.seg_trainer.sg_logger._local_dir, log_level="warning", country_iso_code="ITA")
        self.carbon_tracker.start()

    def init(self, params: dict):
        self.seg_trainer = None
        try:
            self.parse_params(params)
            # trainer_class = KDSegTrainer if kd else SegmentationTrainer
            trainer_class = KDEzTrainer if self.kd else EzTrainer
            self.seg_trainer = trainer_class(
                project_name=self.params['experiment']['name'],
                group_name=self.params['experiment']['group'],
                ckpt_root_dir=self.params['experiment']['tracking_dir'] or 'experiments',
            )
            self.dataset = self.seg_trainer.init_dataset \
                    (params['dataset_interface'], dataset_params=deepcopy(self.dataset_params))
            self.seg_trainer.init_model(params, False, None)
            self.seg_trainer.init_loggers({"in_params": params}, deepcopy(self.train_params))
            logger.info(f"Input params: \n\n {dict_to_yaml_string(params)}")
            self.seg_trainer.print_model_summary()
            self._init_carbon_tracker()
        except Exception as e:
            if (
                self.seg_trainer is not None
                and self.seg_trainer.sg_logger is not None
            ):
                self.seg_trainer.sg_logger.close(really=True, failed=True)
            traceback.print_exception(type(e), e, e.__traceback__)
            raise e

    def resume(self, logger_run: AbstractRunWrapper, updated_config, phases):
        try:
            try:
                self.params = values_to_number(logger_run.get_params())
            except KeyError as e:
                raise RuntimeError("No params recorded for run, just delete it!") from e
            self.params = nested_dict_update(self.params, updated_config)
            self.phases = phases
            logger_run.update_params(self.params)
            self.train_params, self.test_params, self.dataset_params, callbacks, kd = parse_params(self.params)
            self.train_callbacks, self.val_callbacks, self.test_callbacks = callbacks

            # trainer_class = KDSegTrainer if kd else SegmentationTrainer
            trainer_class = KDEzTrainer if kd else EzTrainer
            self.seg_trainer = trainer_class(
                project_name=self.params['experiment']['name'],
                group_name=self.params['experiment']['group'],
                ckpt_root_dir=self.params['experiment']['tracking_dir'] or 'experiments',
            )
            self.dataset = self.seg_trainer.init_dataset \
                (logger_run.get_params()['dataset_interface'], dataset_params=deepcopy(self.dataset_params))
            checkpoint_path = logger_run.get_local_checkpoint_path(phases)
            self.seg_trainer.init_model(self.params, True, checkpoint_path)
            self.seg_trainer.init_loggers({"in_params": self.params}, self.train_params, run_id=logger_run.id, logger_run=logger_run)
            self._init_carbon_tracker()
        except Exception as e:
            if self.seg_trainer is not None and self.seg_trainer.sg_logger is not None:
                self.seg_trainer.sg_logger.close(really=True, failed=True)
            traceback.print_exception(type(e), e, e.__traceback__)
            raise e

    def launch(self):
        try:
            if 'train' in self.phases:
                train(self.seg_trainer, self.train_params, self.dataset, self.train_callbacks, self.val_callbacks)

            if 'test' in self.phases:
                test_metrics = self.seg_trainer.test(**self.test_params, test_phase_callbacks=self.test_callbacks)

            if 'inference' in self.phases:
                inference(self.seg_trainer, self.run_params, self.dataset)
        finally:
            if self.seg_trainer is not None and self.seg_trainer.sg_logger is not None:
                self.upload_emissions()
                self.seg_trainer.sg_logger.close(True)
                
    def upload_emissions(self):
        self.carbon_tracker.stop()
        self.seg_trainer.sg_logger.add_summary({"emissions": self.carbon_tracker.final_emissions})
        emissions_data = self.carbon_tracker.final_emissions_data.values
        emissions_data_values = [[v for k, v in emissions_data.items()]]
        self.seg_trainer.sg_logger.add_table("emissions", emissions_data_values, emissions_data.keys(), [0])

    @property
    def name(self):
        return self.seg_trainer.sg_logger.name

    @property
    def url(self):
        return self.seg_trainer.sg_logger.url


def train(seg_trainer, train_params, dataset, train_callbacks, val_callbacks):
    # Callbacks

    cbcks = [callback_factory(name, params, seg_trainer=seg_trainer, dataset=dataset, loader=dataset.train_loader)
             for name, params in {**train_callbacks, **val_callbacks}.items()
             ]
    # cbcks = [
    #     MetricsLogCallback(Phase.TRAIN_EPOCH_END, freq=1),
    #     MetricsLogCallback(Phase.VALIDATION_EPOCH_END, freq=1),
    #     *cbcks
    # ]
    train_params["phase_callbacks"] = cbcks

    seg_trainer.train(training_params=train_params)
    gc.collect()


def inference(seg_trainer, run_params, dataset):
    run_loader = dataset.get_run_loader(folders=run_params['run_folders'], batch_size=run_params['batch_size'])
    cbcks = [
        # SaveSegmentationPredictionsCallback(phase=Phase.POST_TRAINING,
        #                                     path=
        #                                     run_params['prediction_folder']
        #                                     if run_params['prediction_folder'] != 'mlflow'
        #                                     else mlclient.run.info.artifact_uri + '/predictions',
        #                                     num_classes=len(seg_trainer.test_loader.dataset.classes),
        #                                     )
    ]
    run_loader.dataset.return_name = True
    seg_trainer.run(run_loader, callbacks=cbcks)
    # seg_trainer.valid_loader.dataset.return_name = True
    # seg_trainer.run(seg_trainer.valid_loader, callbacks=cbcks)



from typing import Tuple
import numpy as np
from super_gradients.training.utils.callbacks.all_callbacks import LR_SCHEDULERS_CLS_DICT
import torch
from super_gradients.training.utils.callbacks import Phase
from super_gradients.training.utils.early_stopping import EarlyStop

from ezdl import scheduler as schedulers
from ezdl.loss import instiantiate_loss
from ezdl.metrics import metrics_factory
from ezdl.utils.utilities import recursive_get


def parse_params(params: dict) -> Tuple[dict, dict, dict, Tuple, dict]:
    # Set Random seeds
    torch.manual_seed(params['train_params']['seed'])
    np.random.seed(params['train_params']['seed'])

    # Instantiate loss
    input_train_params = params['train_params']
    loss_params = params['train_params']['loss']
    loss = instiantiate_loss(loss_params['name'], loss_params['params'])

    if "kd" in params:
        if "loss" in params["kd"]:
            loss = init_composed_loss(loss, params["kd"]["loss"])

    if "aux_loss" in params:
        loss = init_composed_loss(loss, params["aux_loss"])

    # metrics
    train_metrics = metrics_factory(params['train_metrics'])
    test_metrics = metrics_factory(params['test_metrics'])

    # dataset params
    dataset_params = params['dataset']

    train_params = {
        **input_train_params,
        "train_metrics_list": list(train_metrics.values()),
        "valid_metrics_list": list(test_metrics.values()),
        "loss": loss,
        "loss_logging_items_names": ["loss"],
        "sg_logger": params['experiment']['logger'],
        'sg_logger_params': {
            'entity': params['experiment']['entity'],
            'tags': params['tags'],
            'project_name': params['experiment']['name'],
        }
    }

    train_params = parse_scheduler(train_params)

    test_params = {
        "test_metrics": test_metrics,
    }

    # callbacks
    train_callbacks = add_phase_in_callbacks(params.get('train_callbacks') or {}, "train")
    test_callbacks = add_phase_in_callbacks(params.get('test_callbacks') or {}, "test")
    val_callbacks = add_phase_in_callbacks(params.get('val_callbacks') or {}, "validation")
        
    # metric to watch
    mwatch = train_params.get('metric_to_watch')
    if mwatch == 'loss' or mwatch.split('/')[0] == 'loss':
        if hasattr(loss, 'component_names'):
            if len(mwatch.split('/')) >= 2:
                mwatch = f'{loss.__class__.__name__}/{"/".join(mwatch.split("/")[1:])}'
            else:
                mwatch = f'{loss.__class__.__name__}/{loss.__class__.__name__}'
        else:
            mwatch = loss.__class__.__name__
        train_params['metric_to_watch'] = mwatch
        
    # early stopping
    early_stopping = None
    if params.get('early_stopping'):
        early_stopping = params.get('early_stopping')
    elif train_params.get('early_stopping_patience'):
        early_stopping = {'patience': train_params['early_stopping_patience']}
    
    if early_stopping:
        val_callbacks['early_stopping'] = early_stopping
        val_callbacks['early_stopping']['monitor'] = mwatch
        val_callbacks['early_stopping']['mode'] = 'max' if train_params['greater_metric_to_watch_is_better'] else 'min'


    # knowledge distillation
    kd = params.get("kd")

    return train_params, test_params, dataset_params, (train_callbacks, val_callbacks, test_callbacks), kd


def init_composed_loss(loss, loss_params):
    """
    Initialize composed loss (e.g. knowledge distillation) and its components
    :param loss: task_loss_function
    :param loss_params: dict of loss parameters
    :return: loss function
    """
    loss_name = loss_params["name"]
    loss_params = loss_params["params"]
    losses_types = [loss_type for loss_type in loss_params if loss_type.endswith("_loss")]
    for loss_type in losses_types:
        loss_type_params = loss_params[loss_type].get("params") or {}
        loss_type_name = loss_params.pop(loss_type)['name'] # remove loss type from loss_params
        loss_params[f"{loss_type}_fn"] = instiantiate_loss(loss_type_name, loss_type_params)
    return instiantiate_loss(loss_name, {**loss_params, "task_loss_fn": loss})


def add_phase_in_callbacks(callbacks, phase):
    """
    Add default phase to callbacks
    :param callbacks: dict of callbacks
    :param phase: "train", "validation" or "test"
    :return: dict of callbacks with phase
    """
    for callback in callbacks.values():
        if callback.get('phase') is None:
            callback['phase'] = phase
    return callbacks


def parse_scheduler(train_params: dict) -> dict:
    """
    Parse scheduler parameters
    """
    scheduler = train_params.get('scheduler') or train_params.get('lr_mode')
    if scheduler is None:
        return train_params
    if "name" in scheduler:
        params = scheduler.get("params") or {}
        scheduler = scheduler["name"]
    if scheduler in LR_SCHEDULERS_CLS_DICT:
        return train_params
    if scheduler in schedulers.__dict__:
        scheduler = schedulers.__dict__[scheduler](**params)
        train_params['lr_mode'] = "function"
        train_params['lr_schedule_function'] = scheduler.perform_scheduling
    return train_params
        

    



import torch
from ezdl.models import WrappedModel

def unwrap_model_from_parallel(model, return_was_wrapped=False):
    """
    Unwrap a model from a DataParallel or DistributedDataParallel wrapper
    :param model: the model
    :return: the unwrapped model
    """
    if isinstance(
        model,
        (
            torch.nn.DataParallel,
            torch.nn.parallel.DistributedDataParallel,
            WrappedModel,
        ),
    ):
        if return_was_wrapped:
            return model.module, True
        return model.module
    else:
        if return_was_wrapped:
            return model, False
        return model


from typing import Mapping
from functools import reduce

from torch import Tensor
from torchmetrics import JaccardIndex, AUROC, F1Score, ConfusionMatrix
from torchmetrics import Precision as TPrecision
from torchmetrics import Recall as TRecall
from torchmetrics import JaccardIndex as TJaccardIndex
from torchmetrics.functional.classification.roc import _roc_compute
import torch

from copy import deepcopy


class AUC(AUROC):
    def update(self, preds: Tensor, target: Tensor) -> None:
        AUROC.update(self, preds.cpu(), target.cpu())

    def get_roc(self):
        preds = torch.cat(self.preds, dim=0)
        target = torch.cat(self.target, dim=0)
        if not self.num_classes:
            raise ValueError(f"`num_classes` bas to be positive number, but got {self.num_classes}")
        return _roc_compute(preds, target, self.num_classes, self.pos_label)


def PerClassAUC(name, code):
    def __init__(self, name, code, *args, **kwargs):
        AUC.__init__(self, **kwargs)
        self.code = code

    def update(self, preds: Tensor, target: Tensor) -> None:
        preds = preds[:, self.code, ::].flatten()
        target = (target == code).flatten()
        AUC.update(self, preds, target)

    metric = type(name, (AUC,), {"update": update, "__init__": __init__})
    return metric(name, code)


def get_multiclass(names, values):
        macro = values.mean()
        per_class = {f"{names[i]}": f1 for i, f1 in enumerate(values)}
        per_class[names[-1]] = macro
        return per_class


def remove_padding(preds, target, padding):
    for i in range(preds.shape[0]):
        w_slice = slice(0, preds.shape[2] - padding[i][1])
        h_slice = slice(0, preds.shape[3] - padding[i][0])
        pred = preds[i, :, w_slice, h_slice]
        targ = target[i, w_slice, h_slice]
        return pred.unsqueeze(0), targ.unsqueeze(0)

class F1(F1Score):
    def __init__(self, *args, **kwargs):
        self.component_names = [f"f1_{i}" for i in range(kwargs['num_classes'])] + ['f1']
        super().__init__(**kwargs, average="none", mdmc_average="global")
        
    def update(self, preds: Tensor, target: Tensor, padding=None) -> None:
        if padding is not None:
            preds, target = remove_padding(preds, target, padding)
        super().update(preds, target)
    
    def compute(self):
        per_class_f1 = super().compute()
        return get_multiclass(self.component_names, per_class_f1)
        
class Precision(TPrecision):
    def __init__(self, *args, **kwargs):
        self.component_names = [f"precision_{i}" for i in range(kwargs['num_classes'])] + ['precision']
        super().__init__(**kwargs, average="none", mdmc_average="global")
        
    def update(self, preds: Tensor, target: Tensor, padding=None) -> None:
        if padding is not None:
            preds, target = remove_padding(preds, target, padding)
        super().update(preds, target)
        
    def compute(self):
        per_class_precision = super().compute()
        return get_multiclass(self.component_names, per_class_precision)
    
class Recall(TRecall):
    def __init__(self, *args, **kwargs):
        self.component_names = [f"recall_{i}" for i in range(kwargs['num_classes'])] + ['recall']
        super().__init__(**kwargs, average="none", mdmc_average="global")
        
    def update(self, preds: Tensor, target: Tensor, padding=None) -> None:
        if padding is not None:
            preds, target = remove_padding(preds, target, padding)
        super().update(preds, target)
    
    def compute(self):
        per_class_recall = super().compute()
        return get_multiclass(self.component_names, per_class_recall)
    

class ConfMat(ConfusionMatrix):
    def __init__(self, num_classes: int, *args, **kwargs):
        self.component_names = [f"cf_{i}_{j}" for i in range(num_classes) for j in range(num_classes)]
        super().__init__(num_classes, *args, **kwargs)
            
    # def update(self, preds: Tensor, target: Tensor) -> None:
    #     return super().update(preds.argmax(dim=1), target.argmax(dim=1))

    def compute(self) -> Tensor:
        # PlaceHolder value
        cf = super(ConfMat, self).compute()
        names = [f"cf_{i}_{j}" for i in range(self.num_classes) for j in range(self.num_classes)]
        return dict(zip(names, map(lambda x: x.item(), cf.flatten())))
    
    def update(self, preds: Tensor, target: Tensor, padding=None) -> None:
        if padding is not None:
            preds, target = remove_padding(preds, target, padding)
        return super().update(preds, target)

    def get_cf(self):
        return super().compute()


def metric_instance(name: str, params: dict) -> dict:
    if params.get('discriminator') is not None:
        params = deepcopy(params)
        names = params.pop('discriminator')
        return {
            subname: METRICS[name](subname, code, **params)
            for subname, code in names
        }
    return {name: METRICS[name](**params)}


def metrics_factory(metrics_params: Mapping) -> dict:
    return reduce(lambda a, b: {**a, **b},
                  [
                      metric_instance(name, params)
                      for name, params in metrics_params.items()
                  ]
                  )
    

def get_metric_titles_components_mapping(metrics):
    m_dict = {
        metric_name:
        metric.component_names if hasattr(metric, "component_names") else [metric_name]
        for metric_name, metric in metrics.items()
    }
    new_dict = {}

    for key, values in m_dict.items():
        for value in values:
            new_dict[value] = key
    return new_dict  


METRICS = {
    'jaccard': JaccardIndex,
    'auc': AUC,
    'perclassauc': PerClassAUC,
    'f1': F1,
    'f1score': F1Score,
    'precision': Precision,
    'recall': Recall,
    'conf_mat': ConfMat,
}



import os

def get_checkpoints_dir_path(project_name: str, group_name: str, ckpt_root_dir: str = None):
    """Creating the checkpoint directory of a given experiment.
    :param experiment_name:     Name of the experiment.
    :param ckpt_root_dir:       Local root directory path where all experiment logging directories will
                                reside. When none is give, it is assumed that pkg_resources.resource_filename('checkpoints', "")
                                exists and will be used.
    :return:                    checkpoints_dir_path
    """
    if ckpt_root_dir:
        return os.path.join(ckpt_root_dir, project_name, group_name)